"""Uart module."""

from __future__ import annotations

import asyncio
import binascii
import logging
from typing import Any, Callable

import zigpy.config
import zigpy.serial

LOGGER = logging.getLogger(__name__)


class Gateway(zigpy.serial.SerialProtocol):
    END = b"\xC0"
    ESC = b"\xDB"
    ESC_END = b"\xDC"
    ESC_ESC = b"\xDD"

    def __init__(self, api):
        """Initialize instance of the UART gateway."""
        super().__init__()
        self._api = api

    def connection_lost(self, exc: Exception | None) -> None:
        """Port was closed expectedly or unexpectedly."""
        super().connection_lost(exc)

        if self._api is not None:
            self._api.connection_lost(exc)

    def close(self):
        super().close()
        self._api = None

    def send(self, data: bytes) -> None:
        """Send data, taking care of escaping and framing."""
        checksum = bytes(self._checksum(data))
        frame = self._escape(data + checksum)
        self._transport.write(self.END + frame + self.END)

    def data_received(self, data: bytes) -> None:
        """Handle data received from the uart."""
        super().data_received(data)

        while self._buffer:
            end = self._buffer.find(self.END)
            if end < 0:
                return None

            frame = self._buffer[:end]
            self._buffer = self._buffer[(end + 1) :]
            frame = self._unescape(frame)

            if len(frame) < 4:
                continue

            checksum = frame[-2:]
            frame = frame[:-2]
            if self._checksum(frame) != checksum:
                LOGGER.warning(
                    "Invalid checksum: 0x%s, data: 0x%s",
                    binascii.hexlify(checksum).decode(),
                    binascii.hexlify(frame).decode(),
                )
                continue

            LOGGER.debug("Frame received: 0x%s", binascii.hexlify(frame).decode())
            try:
                self._api.data_received(frame)
            except Exception as exc:
                LOGGER.error("Unexpected error handling the frame", exc_info=exc)

    def _unescape(self, data):
        ret = []
        idx = 0
        while idx < len(data):
            b = data[idx]
            if b == self.ESC[0]:
                idx += 1
                if idx >= len(data):
                    return None
                elif data[idx] == self.ESC_END[0]:
                    b = self.END[0]
                elif data[idx] == self.ESC_ESC[0]:
                    b = self.ESC[0]
            ret.append(b)
            idx += 1

        return bytes(ret)

    def _escape(self, data):
        ret = []
        for b in data:
            if b == self.END[0]:
                ret.append(self.ESC[0])
                ret.append(self.ESC_END[0])
            elif b == self.ESC[0]:
                ret.append(self.ESC[0])
                ret.append(self.ESC_ESC[0])
            else:
                ret.append(b)
        return bytes(ret)

    def _checksum(self, data):
        ret = []
        s = ~(sum(data)) + 1
        ret.append(s % 0x100)
        ret.append((s >> 8) % 0x100)
        return bytes(ret)


async def connect(config: dict[str, Any], api: Callable) -> Gateway:
    protocol = Gateway(api)

    LOGGER.debug("Connecting to %s", config[zigpy.config.CONF_DEVICE_PATH])

    _, protocol = await zigpy.serial.create_serial_connection(
        loop=asyncio.get_running_loop(),
        protocol_factory=lambda: protocol,
        url=config[zigpy.config.CONF_DEVICE_PATH],
        baudrate=config[zigpy.config.CONF_DEVICE_BAUDRATE],
        flow_control=config[zigpy.config.CONF_DEVICE_FLOW_CONTROL],
    )

    await protocol.wait_until_connected()

    return protocol



"""Test sending and receiving Zigbee packets using the zigpy packet API."""

import asyncio
import contextlib
from unittest.mock import patch

import pytest
import zigpy.exceptions
import zigpy.types as zigpy_t

from zigpy_deconz.api import Status, TXStatus
import zigpy_deconz.exception
import zigpy_deconz.types as t

from tests.test_application import api, app, device_path  # noqa: F401


@contextlib.contextmanager
def patch_data_request(app, *, fail_enqueue=False, fail_deliver=False):  # noqa: F811
    with patch.object(app._api, "aps_data_request") as mock_request:

        async def mock_send(req_id, *args, **kwargs):
            await asyncio.sleep(0)

            if fail_enqueue:
                raise zigpy_deconz.exception.CommandError(
                    "Error", status=Status.FAILURE, command=None
                )

            if fail_deliver:
                app.handle_tx_confirm(
                    req_id, TXStatus(int(zigpy_t.APSStatus.APS_NO_ACK))
                )
            else:
                app.handle_tx_confirm(req_id, TXStatus.SUCCESS)

        mock_request.side_effect = mock_send

        yield mock_request


@pytest.fixture
def tx_packet():
    return zigpy_t.ZigbeePacket(
        src=zigpy_t.AddrModeAddress(addr_mode=zigpy_t.AddrMode.NWK, address=0x0000),
        src_ep=0x12,
        dst=zigpy_t.AddrModeAddress(addr_mode=zigpy_t.AddrMode.NWK, address=0x1234),
        dst_ep=0x34,
        tsn=0x56,
        profile_id=0x7890,
        cluster_id=0xABCD,
        data=zigpy_t.SerializableBytes(b"some data"),
        tx_options=zigpy_t.TransmitOptions.ACK,
        radius=0,
    )


async def test_send_packet_nwk(app, tx_packet):  # noqa: F811
    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.NWK
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == (
        t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
        | t.DeconzTransmitOptions.USE_APS_ACKS
    )
    assert req["relays"] is None
    assert req["radius"] == 0


async def test_send_packet_nwk_no_ack(app, tx_packet):  # noqa: F811
    tx_packet.tx_options &= ~zigpy_t.TransmitOptions.ACK

    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.NWK
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
    assert req["relays"] is None
    assert req["radius"] == 0


async def test_send_packet_source_route(app, tx_packet):  # noqa: F811
    tx_packet.source_route = [0xAABB, 0xCCDD]

    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.NWK
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == (
        t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
        | t.DeconzTransmitOptions.USE_APS_ACKS
    )
    assert req["relays"] == [0xAABB, 0xCCDD]
    assert req["radius"] == 0


async def test_send_packet_ieee(app, tx_packet):  # noqa: F811
    tx_packet.dst = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.IEEE,
        address=zigpy_t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
    )

    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.IEEE
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == (
        t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
        | t.DeconzTransmitOptions.USE_APS_ACKS
    )
    assert req["relays"] is None
    assert req["radius"] == 0


async def test_send_packet_group(app, tx_packet):  # noqa: F811
    tx_packet.dst = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.Group, address=0x1234
    )
    tx_packet.radius = 12

    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.GROUP
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
    assert req["relays"] is None
    assert req["radius"] == 12


async def test_send_packet_broadcast(app, tx_packet):  # noqa: F811
    tx_packet.dst = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.Broadcast,
        address=zigpy_t.BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR,
    )
    tx_packet.radius = 12

    with patch_data_request(app) as mock_req:
        await app.send_packet(tx_packet)

    assert len(mock_req.mock_calls) == 1
    req = mock_req.mock_calls[0].kwargs

    assert req["dst_addr_ep"].address_mode == t.AddressMode.NWK
    assert req["dst_addr_ep"].address == tx_packet.dst.address
    assert req["dst_addr_ep"].endpoint == tx_packet.dst_ep
    assert req["profile"] == tx_packet.profile_id
    assert req["cluster"] == tx_packet.cluster_id
    assert req["aps_payload"] == tx_packet.data.serialize()
    assert req["tx_options"] == t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY
    assert req["relays"] is None
    assert req["radius"] == 12


async def test_send_packet_enqueue_failure(app, tx_packet):  # noqa: F811
    with patch_data_request(app, fail_enqueue=True):  # noqa: F811
        with pytest.raises(zigpy.exceptions.DeliveryError) as e:
            await app.send_packet(tx_packet)

    assert "Failed to enqueue" in str(e)


async def test_send_packet_deliver_failure(app, tx_packet):  # noqa: F811
    with patch_data_request(app, fail_deliver=True):  # noqa: F811
        with pytest.raises(zigpy.exceptions.DeliveryError) as e:
            await app.send_packet(tx_packet)

    assert "Failed to deliver" in str(e)



"""ControllerApplication for deCONZ protocol based adapters."""

from __future__ import annotations

import asyncio
import importlib.metadata
import logging
import sys
from typing import Any

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

import zigpy.application
import zigpy.config
import zigpy.device
import zigpy.endpoint
import zigpy.exceptions
from zigpy.exceptions import FormationFailure, NetworkNotFormed
import zigpy.state
import zigpy.types
import zigpy.util
import zigpy.zdo.types as zdo_t

import zigpy_deconz
from zigpy_deconz import types as t
from zigpy_deconz.api import (
    Deconz,
    FirmwarePlatform,
    IndexedEndpoint,
    IndexedKey,
    LinkKey,
    NetworkParameter,
    NetworkState,
    SecurityMode,
    Status,
    TXStatus,
)
from zigpy_deconz.config import CONFIG_SCHEMA
import zigpy_deconz.exception
from zigpy_deconz.utils import is_usb_serial_port

LIB_VERSION = importlib.metadata.version("zigpy-deconz")
LOGGER = logging.getLogger(__name__)

CHANGE_NETWORK_POLL_TIME = 1
CHANGE_NETWORK_STATE_DELAY = 2
DELAY_NEIGHBOUR_SCAN_S = 1500
SEND_CONFIRM_TIMEOUT = 60

PROTO_VER_MANUAL_SOURCE_ROUTE = 0x010C
PROTO_VER_WATCHDOG = 0x0108
PROTO_VER_NEIGBOURS = 0x0107

CONBEE_III_ENERGY_SCAN_ATTEMPTS = 5


class ControllerApplication(zigpy.application.ControllerApplication):
    SCHEMA = CONFIG_SCHEMA

    _probe_config_variants = [
        {zigpy.config.CONF_DEVICE_BAUDRATE: 38400},
        {zigpy.config.CONF_DEVICE_BAUDRATE: 115200},
    ]

    _watchdog_period = 30

    def __init__(self, config: dict[str, Any]):
        """Initialize instance."""

        super().__init__(config=config)
        self._api = None

        self._pending_requests = {}

        self._delayed_neighbor_scan_task = None
        self._reconnect_task = None

        self._written_endpoints = set()

    async def _watchdog_feed(self):
        if self._api.protocol_version >= PROTO_VER_WATCHDOG and not (
            self._api.firmware_version.platform == FirmwarePlatform.Conbee_III
            and self._api.firmware_version <= 0x26450900
        ):
            await self._api.write_parameter(
                NetworkParameter.watchdog_ttl, int(2 * self._watchdog_period)
            )
        else:
            await self._api.get_device_state()

    async def connect(self):
        api = Deconz(self, self._config[zigpy.config.CONF_DEVICE])

        try:
            await api.connect()
        except Exception:
            await api.disconnect()
            raise

        self._api = api
        self._written_endpoints.clear()

    async def disconnect(self):
        if self._delayed_neighbor_scan_task is not None:
            self._delayed_neighbor_scan_task.cancel()
            self._delayed_neighbor_scan_task = None

        if self._api is not None:
            await self._api.disconnect()
            self._api = None

    async def permit_with_link_key(self, node: t.EUI64, link_key: t.KeyData, time_s=60):
        await self._api.write_parameter(
            NetworkParameter.link_key,
            LinkKey(ieee=node, key=link_key),
        )
        await self.permit(time_s)

    async def start_network(self):
        await self.register_endpoints()
        await self.load_network_info(load_devices=False)
        await self._change_network_state(NetworkState.CONNECTED)

        coordinator = await DeconzDevice.new(
            self,
            self.state.node_info.ieee,
            self.state.node_info.nwk,
            self.state.node_info.model,
        )

        self.devices[self.state.node_info.ieee] = coordinator
        if self._api.protocol_version >= PROTO_VER_NEIGBOURS and not (
            self._api.firmware_version.platform == FirmwarePlatform.Conbee_III
            and self._api.firmware_version < 0x264D0900
        ):
            await self.restore_neighbours()

        self._delayed_neighbor_scan_task = asyncio.create_task(
            self._delayed_neighbour_scan()
        )

    async def _change_network_state(
        self,
        target_state: NetworkState,
        *,
        timeout: int = 10 * CHANGE_NETWORK_POLL_TIME,
    ):
        async def change_loop():
            while True:
                try:
                    device_state = await self._api.get_device_state()
                except asyncio.TimeoutError:
                    # 0x264B0900 and earlier can reset during device state changes
                    # requiring a firmware reset, causing state polling to fail
                    LOGGER.debug("Failed to poll device state")
                else:
                    if NetworkState(device_state.network_state) == target_state:
                        break

                await asyncio.sleep(CHANGE_NETWORK_POLL_TIME)

        await self._api.change_network_state(target_state)

        try:
            async with asyncio_timeout(timeout):
                await change_loop()
        except asyncio.TimeoutError:
            if target_state != NetworkState.CONNECTED:
                raise

            raise FormationFailure(
                "Network formation refused: there is likely too much RF interference."
                " Make sure your coordinator is on a USB 2.0 extension cable and"
                " away from any sources of interference, like USB 3.0 ports, SSDs,"
                " 2.4GHz routers, motherboards, etc."
            )

    async def reset_network_info(self):
        # TODO: There does not appear to be a way to factory reset a Conbee
        await self.write_network_info(
            network_info=zigpy.state.NetworkInfo(
                pan_id=0xFFFF,
                extended_pan_id=zigpy.types.EUI64.convert("FF:FF:FF:FF:FF:FF:FF:FF"),
                channel=None,
                channel_mask=zigpy.types.Channels(0),
                nwk_update_id=0,
                network_key=zigpy.state.Key(
                    key=zigpy.types.KeyData.convert(
                        "FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF:FF"
                    )
                ),
                tc_link_key=zigpy.state.Key(
                    key=zigpy.types.KeyData.convert(b"ZigBeeAlliance09".hex())
                ),
                security_level=0x05,
            ),
            node_info=zigpy.state.NodeInfo(
                logical_type=zdo_t.LogicalType.Coordinator,
                ieee=zigpy.types.EUI64.UNKNOWN,
                nwk=0xFFFF,
            ),
        )

    async def write_network_info(self, *, network_info, node_info):
        try:
            await self._api.write_parameter(
                NetworkParameter.nwk_frame_counter, network_info.network_key.tx_counter
            )
        except zigpy_deconz.exception.CommandError as ex:
            assert ex.status == Status.UNSUPPORTED

            # If we are resetting the adapter or forming a brand new network, we can
            # skip this check
            if not (
                network_info.stack_specific.get("form_quickly", False)
                or network_info.network_key.tx_counter == 0
            ):
                fw_version = f"{int(self._api.firmware_version):#010x}"
                raise zigpy.exceptions.CannotWriteNetworkSettings(
                    f"Please upgrade your adapter firmware. Firmware version"
                    f" {fw_version} does not support writing the network key frame"
                    f" counter, which is required for migration to succeed."
                )

        if node_info.logical_type == zdo_t.LogicalType.Coordinator:
            await self._api.write_parameter(
                NetworkParameter.aps_designed_coordinator, 1
            )
        else:
            await self._api.write_parameter(
                NetworkParameter.aps_designed_coordinator, 0
            )

        await self._api.write_parameter(NetworkParameter.nwk_address, node_info.nwk)

        if node_info.ieee != zigpy.types.EUI64.UNKNOWN:
            # TODO: is there a way to revert it back to the hardware default? Or is this
            #       information lost when the parameter is overwritten?
            await self._api.write_parameter(
                NetworkParameter.mac_address, node_info.ieee
            )
            node_ieee = node_info.ieee
        else:
            ieee = await self._api.read_parameter(NetworkParameter.mac_address)
            node_ieee = zigpy.types.EUI64(ieee)

        # There is no way to specify both a mask and the logical channel
        if network_info.channel is not None:
            channel_mask = zigpy.types.Channels.from_channel_list(
                [network_info.channel]
            )

            if network_info.channel_mask and channel_mask != network_info.channel_mask:
                LOGGER.warning(
                    "Channel mask %s will be replaced with current logical channel %s",
                    network_info.channel_mask,
                    channel_mask,
                )
        else:
            channel_mask = network_info.channel_mask

        await self._api.write_parameter(NetworkParameter.channel_mask, channel_mask)
        await self._api.write_parameter(NetworkParameter.use_predefined_nwk_panid, True)
        await self._api.write_parameter(NetworkParameter.nwk_panid, network_info.pan_id)
        await self._api.write_parameter(
            NetworkParameter.aps_extended_panid, network_info.extended_pan_id
        )
        await self._api.write_parameter(
            NetworkParameter.nwk_update_id, network_info.nwk_update_id
        )

        await self._api.write_parameter(
            NetworkParameter.network_key,
            IndexedKey(index=0, key=network_info.network_key.key),
        )

        if network_info.network_key.seq != 0:
            LOGGER.warning(
                "Non-zero network key sequence number is not supported: %s",
                network_info.network_key.seq,
            )

        tc_link_key_partner_ieee = network_info.tc_link_key.partner_ieee

        if tc_link_key_partner_ieee == zigpy.types.EUI64.UNKNOWN:
            tc_link_key_partner_ieee = node_ieee

        await self._api.write_parameter(
            NetworkParameter.trust_center_address,
            tc_link_key_partner_ieee,
        )
        await self._api.write_parameter(
            NetworkParameter.link_key,
            LinkKey(
                ieee=tc_link_key_partner_ieee,
                key=network_info.tc_link_key.key,
            ),
        )

        if self._api.firmware_version.platform != FirmwarePlatform.Conbee_III:
            if network_info.security_level == 0x00:
                await self._api.write_parameter(
                    NetworkParameter.security_mode, SecurityMode.NO_SECURITY
                )
            else:
                await self._api.write_parameter(
                    NetworkParameter.security_mode, SecurityMode.ONLY_TCLK
                )

        # Note: Changed network configuration parameters become only affective after
        # sending a Leave Network Request followed by a Create or Join Network Request
        await self._change_network_state(NetworkState.OFFLINE)

        if (
            network_info.pan_id == 0xFFFF
            or network_info.channel_mask == zigpy.types.Channels(0)
        ):
            # Network is being reset, it will never enter the CONNECTED state
            return

        await asyncio.sleep(CHANGE_NETWORK_STATE_DELAY)
        await self._change_network_state(NetworkState.CONNECTED)

    async def load_network_info(self, *, load_devices=False):
        network_info = self.state.network_info
        node_info = self.state.node_info

        ieee = await self._api.read_parameter(NetworkParameter.mac_address)
        node_info.ieee = zigpy.types.EUI64(ieee)
        designed_coord = await self._api.read_parameter(
            NetworkParameter.aps_designed_coordinator
        )

        if designed_coord == 0x01:
            node_info.logical_type = zdo_t.LogicalType.Coordinator
        else:
            node_info.logical_type = zdo_t.LogicalType.Router

        node_info.nwk = await self._api.read_parameter(NetworkParameter.nwk_address)

        node_info.manufacturer = "dresden elektronik"

        is_usb = await asyncio.get_running_loop().run_in_executor(
            None,
            is_usb_serial_port,
            self._config[zigpy.config.CONF_DEVICE][zigpy.config.CONF_DEVICE_PATH],
        )

        if is_usb:
            node_info.model = "Conbee"
        else:
            node_info.model = "Raspbee"

        node_info.model += {
            FirmwarePlatform.Conbee: "",
            FirmwarePlatform.Conbee_II: " II",
            FirmwarePlatform.Conbee_III: " III",
        }[self._api.firmware_version.platform]

        node_info.version = f"{int(self._api.firmware_version):#010x}"

        network_info.source = f"zigpy-deconz@{LIB_VERSION}"
        network_info.metadata = {
            "deconz": {
                "version": node_info.version,
            }
        }

        network_info.pan_id = await self._api.read_parameter(NetworkParameter.nwk_panid)
        network_info.extended_pan_id = await self._api.read_parameter(
            NetworkParameter.aps_extended_panid
        )

        if network_info.extended_pan_id == zigpy.types.ExtendedPanId.convert(
            "00:00:00:00:00:00:00:00"
        ):
            network_info.extended_pan_id = await self._api.read_parameter(
                NetworkParameter.nwk_extended_panid
            )

        network_info.channel = await self._api.read_parameter(
            NetworkParameter.current_channel
        )
        network_info.channel_mask = await self._api.read_parameter(
            NetworkParameter.channel_mask
        )
        network_info.nwk_update_id = await self._api.read_parameter(
            NetworkParameter.nwk_update_id
        )

        if (
            node_info.nwk == 0xFFFF
            or network_info.pan_id == 0xFFFF
            or (
                network_info.extended_pan_id
                == zigpy.types.ExtendedPanId.convert("FF:FF:FF:FF:FF:FF:FF:FF")
            )
            or network_info.channel == 0
        ):
            raise NetworkNotFormed("Network is not formed")

        indexed_key = await self._api.read_parameter(NetworkParameter.network_key, 0)

        network_info.network_key = zigpy.state.Key()
        network_info.network_key.key = indexed_key.key

        try:
            network_info.network_key.tx_counter = await self._api.read_parameter(
                NetworkParameter.nwk_frame_counter
            )
        except zigpy_deconz.exception.CommandError as ex:
            assert ex.status == Status.UNSUPPORTED
            LOGGER.debug(
                "Conbee firmware is too old and does not support reading the frame"
                " counter, picking a large value instead"
            )
            network_info.network_key.tx_counter = 2**30

        network_info.tc_link_key = zigpy.state.Key()
        network_info.tc_link_key.partner_ieee = await self._api.read_parameter(
            NetworkParameter.trust_center_address
        )

        link_key = await self._api.read_parameter(
            NetworkParameter.link_key,
            network_info.tc_link_key.partner_ieee,
        )
        network_info.tc_link_key.key = link_key.key

        security_mode = await self._api.read_parameter(NetworkParameter.security_mode)

        if security_mode == SecurityMode.NO_SECURITY:
            network_info.security_level = 0x00
        elif security_mode == SecurityMode.ONLY_TCLK:
            network_info.security_level = 0x05
        else:
            LOGGER.warning("Unsupported security mode %r", security_mode)
            network_info.security_level = 0x05

    async def force_remove(self, dev):
        """Forcibly remove device from NCP."""

    async def energy_scan(
        self, channels: t.Channels.ALL_CHANNELS, duration_exp: int, count: int
    ) -> dict[int, float]:
        if self._api.firmware_version.platform in (
            FirmwarePlatform.Conbee,
            FirmwarePlatform.Conbee_II,
        ):
            results = await super().energy_scan(
                channels=channels, duration_exp=duration_exp, count=count
            )

            # The Conbee I/II seems to max out at an LQI of 85, which is exactly 255/3
            return {c: v * 3 for c, v in results.items()}

        for i in range(CONBEE_III_ENERGY_SCAN_ATTEMPTS):
            # The Conbee III energy scan inherits the EmberZNet ZDO bug
            try:
                rsp = await self._device.zdo.Mgmt_NWK_Update_req(
                    zigpy.zdo.types.NwkUpdate(
                        ScanChannels=channels,
                        ScanDuration=duration_exp,
                        ScanCount=count,
                    )
                )
                break
            except (asyncio.TimeoutError, zigpy.exceptions.DeliveryError):
                if i == CONBEE_III_ENERGY_SCAN_ATTEMPTS - 1:
                    raise

                continue

        _, scanned_channels, _, _, energy_values = rsp
        return dict(zip(scanned_channels, energy_values))

    async def _move_network_to_channel(
        self, new_channel: int, new_nwk_update_id: int
    ) -> None:
        """Move device to a new channel."""
        channel_mask = zigpy.types.Channels.from_channel_list([new_channel])
        await self._api.write_parameter(NetworkParameter.channel_mask, channel_mask)
        await self._api.write_parameter(
            NetworkParameter.nwk_update_id, new_nwk_update_id
        )

        await self._change_network_state(NetworkState.OFFLINE)
        await asyncio.sleep(CHANGE_NETWORK_STATE_DELAY)
        await self._change_network_state(NetworkState.CONNECTED)

    async def add_endpoint(self, descriptor: zdo_t.SimpleDescriptor) -> None:
        """Register an endpoint on the device, replacing any with conflicting IDs."""

        endpoints = {}

        # Read and count the current endpoints. Some firmwares have three, others four.
        for index in range(255 + 1):
            try:
                current_descriptor = await self._api.read_parameter(
                    NetworkParameter.configure_endpoint, index
                )
            except zigpy_deconz.exception.CommandError as ex:
                assert ex.status == Status.UNSUPPORTED
                break
            else:
                endpoints[index] = current_descriptor.descriptor

        LOGGER.debug("Got endpoint slots: %r", endpoints)

        # Don't write endpoints unnecessarily
        if descriptor in endpoints.values():
            LOGGER.debug("Endpoint already registered, skipping")

            # Pretend we wrote it
            self._written_endpoints.add(list(endpoints.values()).index(descriptor))
            return

        # Keep track of the best endpoint descriptor to replace
        target_index = None

        for index, current_descriptor in endpoints.items():
            # Ignore ones we've already written
            if index in self._written_endpoints:
                continue

            target_index = index

            if current_descriptor.endpoint == descriptor.endpoint:
                # Prefer to replace the endpoint with the same ID
                break

        if target_index is None:
            raise ValueError(f"No available endpoint slots exist: {endpoints!r}")

        LOGGER.debug("Writing %s to slot %r", descriptor, target_index)

        await self._api.write_parameter(
            NetworkParameter.configure_endpoint,
            IndexedEndpoint(index=target_index, descriptor=descriptor),
        )

    async def send_packet(self, packet):
        LOGGER.debug("Sending packet: %r", packet)

        tx_options = t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY

        if (
            zigpy.types.TransmitOptions.ACK in packet.tx_options
            and packet.dst.addr_mode
            in (zigpy.types.AddrMode.NWK, zigpy.types.AddrMode.IEEE)
        ):
            tx_options |= t.DeconzTransmitOptions.USE_APS_ACKS

        async with self._limit_concurrency(priority=packet.priority):
            req_id = self.get_sequence()

            if req_id in self._pending_requests:
                raise zigpy.exceptions.DeliveryError(
                    f"Request with id {req_id} is already pending, cannot send"
                )

            future = self._pending_requests[req_id] = asyncio.Future()

            try:
                try:
                    await self._api.aps_data_request(
                        req_id=req_id,
                        dst_addr_ep=t.DeconzAddressEndpoint.from_zigpy_type(
                            packet.dst, packet.dst_ep or 0
                        ),
                        profile=packet.profile_id,
                        cluster=packet.cluster_id,
                        src_ep=min(1, packet.src_ep),
                        aps_payload=packet.data.serialize(),
                        tx_options=tx_options,
                        relays=packet.source_route,
                        radius=packet.radius or 0,
                    )
                except zigpy_deconz.exception.CommandError as ex:
                    raise zigpy.exceptions.DeliveryError(
                        f"Failed to enqueue packet: {ex!r}", ex.status
                    )

                async with asyncio_timeout(SEND_CONFIRM_TIMEOUT):
                    status = await future

                if status != TXStatus.SUCCESS:
                    raise zigpy.exceptions.DeliveryError(
                        f"Failed to deliver packet: {status!r}", status
                    )
            finally:
                del self._pending_requests[req_id]

    async def permit_ncp(self, time_s=60):
        assert 0 <= time_s <= 254
        await self._api.write_parameter(NetworkParameter.permit_join, time_s)

    def handle_tx_confirm(self, req_id, status):
        try:
            future = self._pending_requests[req_id]
        except KeyError:
            LOGGER.warning(
                "Unexpected transmit confirm for request id %s, Status: %s",
                req_id,
                status,
            )
        else:
            try:
                future.set_result(status)
            except asyncio.InvalidStateError as exc:
                LOGGER.debug(
                    "Invalid state on future - probably duplicate response: %s", exc
                )

    async def restore_neighbours(self) -> None:
        """Restore children."""
        coord = self.get_device(ieee=self.state.node_info.ieee)

        for neighbor in self.topology.neighbors[coord.ieee]:
            try:
                device = self.get_device(ieee=neighbor.ieee)
            except KeyError:
                continue

            descr = device.node_desc
            LOGGER.debug(
                "device: 0x%04x - %s %s, FFD=%s, Rx_on_when_idle=%s",
                device.nwk,
                device.manufacturer,
                device.model,
                descr.is_full_function_device if descr is not None else None,
                descr.is_receiver_on_when_idle if descr is not None else None,
            )
            if (
                descr is None
                or descr.is_full_function_device
                or descr.is_receiver_on_when_idle
            ):
                continue

            LOGGER.debug("Restoring %s as direct child", device)

            try:
                await self._api.add_neighbour(
                    nwk=device.nwk,
                    ieee=device.ieee,
                    mac_capability_flags=descr.mac_capability_flags,
                )
            except zigpy_deconz.exception.CommandError as ex:
                assert ex.status == Status.FAILURE
                LOGGER.debug("Failed to add device to neighbor table: %s", ex)

    async def _delayed_neighbour_scan(self) -> None:
        """Scan coordinator's neighbours."""
        await asyncio.sleep(DELAY_NEIGHBOUR_SCAN_S)
        coord = self.get_device(ieee=self.state.node_info.ieee)
        await self.topology.scan(devices=[coord])


class DeconzDevice(zigpy.device.Device):
    """Zigpy Device representing Coordinator."""

    def __init__(self, model: str, *args, **kwargs):
        """Initialize instance."""

        super().__init__(*args, **kwargs)
        self._model = model

    async def add_to_group(self, grp_id: int, name: str = None) -> None:
        group = self.application.groups.add_group(grp_id, name)

        for epid in self.endpoints:
            if not epid:
                continue  # skip ZDO
            group.add_member(self.endpoints[epid])
        return [0]

    async def remove_from_group(self, grp_id: int) -> None:
        for epid in self.endpoints:
            if not epid:
                continue  # skip ZDO
            self.application.groups[grp_id].remove_member(self.endpoints[epid])
        return [0]

    @property
    def manufacturer(self):
        return "dresden elektronik"

    @property
    def model(self):
        return self._model

    @classmethod
    async def new(cls, application, ieee, nwk, model: str):
        """Create or replace zigpy device."""
        dev = cls(model, application, ieee, nwk)

        if ieee in application.devices:
            from_dev = application.get_device(ieee=ieee)
            dev.status = from_dev.status
            dev.node_desc = from_dev.node_desc
            for ep_id, from_ep in from_dev.endpoints.items():
                if not ep_id:
                    continue  # Skip ZDO
                ep = dev.add_endpoint(ep_id)
                ep.profile_id = from_ep.profile_id
                ep.device_type = from_ep.device_type
                ep.status = from_ep.status
                ep.in_clusters = from_ep.in_clusters
                ep.out_clusters = from_ep.out_clusters
        else:
            application.devices[ieee] = dev
            await dev.initialize()

        return dev



"""Test exceptions."""

from unittest import mock

import zigpy_deconz.exception


def test_command_error():
    ex = zigpy_deconz.exception.CommandError(
        mock.sentinel.message,
        status=mock.sentinel.status,
        command=mock.sentinel.command,
    )
    assert ex.status is mock.sentinel.status
    assert ex.command is mock.sentinel.command



"""Test `load_network_info` and `write_network_info` methods."""

import importlib.metadata
from unittest.mock import ANY, call

import pytest
from zigpy.exceptions import CannotWriteNetworkSettings, NetworkNotFormed
import zigpy.state as app_state
import zigpy.types as t
import zigpy.zdo.types as zdo_t

import zigpy_deconz
import zigpy_deconz.api
import zigpy_deconz.exception
import zigpy_deconz.types
import zigpy_deconz.zigbee.application as application

from tests.async_mock import AsyncMock, patch
from tests.test_application import api, app, device_path  # noqa: F401


def merge_objects(obj: object, update: dict) -> None:
    for key, value in update.items():
        if "." not in key:
            setattr(obj, key, value)
        else:
            subkey, rest = key.split(".", 1)
            merge_objects(getattr(obj, subkey), {rest: value})


@pytest.fixture
def node_info():
    return app_state.NodeInfo(
        nwk=t.NWK(0x0000),
        ieee=t.EUI64.convert("93:2C:A9:34:D9:D0:5D:12"),
        logical_type=zdo_t.LogicalType.Coordinator,
        manufacturer="dresden elektronik",
        model="Conbee II",
        version="0x26580700",
    )


@pytest.fixture
def network_info(node_info):
    return app_state.NetworkInfo(
        extended_pan_id=t.ExtendedPanId.convert("0D:49:91:99:AE:CD:3C:35"),
        pan_id=t.PanId(0x9BB0),
        nwk_update_id=0x12,
        nwk_manager_id=t.NWK(0x0000),
        channel=t.uint8_t(15),
        channel_mask=t.Channels.from_channel_list([15, 20, 25]),
        security_level=t.uint8_t(5),
        network_key=app_state.Key(
            key=t.KeyData.convert("9A:79:D6:9A:DA:EC:45:C6:F2:EF:EB:AF:DA:A3:07:B6"),
            seq=108,
            tx_counter=39009277,
        ),
        tc_link_key=app_state.Key(
            key=t.KeyData(b"ZigBeeAlliance09"),
            partner_ieee=node_info.ieee,
            tx_counter=8712428,
        ),
        key_table=[],
        children=[],
        nwk_addresses={},
        stack_specific={},
        source=f"zigpy-deconz@{importlib.metadata.version('zigpy-deconz')}",
        metadata={"deconz": {"version": "0x26580700"}},
    )


@patch.object(application, "CHANGE_NETWORK_POLL_TIME", 0.001)
@patch.object(application, "CHANGE_NETWORK_STATE_DELAY", 0.001)
@pytest.mark.parametrize(
    (
        "channel_mask",
        "channel",
        "security_level",
        "fw_supports_fc",
        "logical_type",
        "tx_counter",
        "should_error",
    ),
    [
        # FW supports frame counter
        (
            t.Channels.from_channel_list([15]),
            15,
            0,
            True,
            zdo_t.LogicalType.Coordinator,
            39009277,
            False,
        ),
        # FW doesn't support but we're writing blank settings (tx_counter == 0)
        (
            t.Channels.from_channel_list([15]),
            15,
            5,
            False,
            zdo_t.LogicalType.Coordinator,
            0,
            False,
        ),
        # FW doesn't support and we're writing real settings (should error)
        (
            t.Channels.from_channel_list([15]),
            15,
            0,
            False,
            zdo_t.LogicalType.Coordinator,
            39009277,
            True,
        ),
        # Additional test cases with FW support
        (
            t.Channels.from_channel_list([15, 20]),
            15,
            5,
            True,
            zdo_t.LogicalType.Coordinator,
            39009277,
            False,
        ),
        (
            t.Channels.from_channel_list([15, 20, 25]),
            None,
            5,
            True,
            zdo_t.LogicalType.Router,
            39009277,
            False,
        ),
        (
            None,
            15,
            5,
            True,
            zdo_t.LogicalType.Coordinator,
            39009277,
            False,
        ),
    ],
)
async def test_write_network_info(
    app,  # noqa: F811
    network_info,
    node_info,
    channel_mask,
    channel,
    security_level,
    fw_supports_fc,
    logical_type,
    tx_counter,
    should_error,
):
    """Test that network info is correctly written."""

    params = {}

    async def write_parameter(param, *args):
        if (
            not fw_supports_fc
            and param == zigpy_deconz.api.NetworkParameter.nwk_frame_counter
        ):
            raise zigpy_deconz.exception.CommandError(
                "Command is unsupported",
                status=zigpy_deconz.api.Status.UNSUPPORTED,
                command=None,
            )

        params[param.name] = args

    app._change_network_state = AsyncMock()
    app._api.write_parameter = AsyncMock(side_effect=write_parameter)

    network_info = network_info.replace(
        channel=channel,
        channel_mask=channel_mask,
        security_level=security_level,
        network_key=network_info.network_key.replace(tx_counter=tx_counter),
    )

    node_info = node_info.replace(logical_type=logical_type)

    if should_error:
        with pytest.raises(
            CannotWriteNetworkSettings,
            match=(
                "Please upgrade your adapter firmware. Firmware version 0x26580700 does"
                " not support writing the network key frame counter, which is required"
                " for migration to succeed."
            ),
        ):
            await app.write_network_info(
                network_info=network_info,
                node_info=node_info,
            )
        return

    await app.write_network_info(
        network_info=network_info,
        node_info=node_info,
    )

    params = {
        call[0][0].name: call[0][1:]
        for call in app._api.write_parameter.await_args_list
    }

    # Only check frame counter if firmware supports it
    if fw_supports_fc:
        assert params["nwk_frame_counter"] == (network_info.network_key.tx_counter,)

    if node_info.logical_type == zdo_t.LogicalType.Coordinator:
        assert params["aps_designed_coordinator"] == (1,)
    else:
        assert params["aps_designed_coordinator"] == (0,)

    assert params["nwk_address"] == (node_info.nwk,)
    assert params["mac_address"] == (node_info.ieee,)

    if channel is not None:
        assert params["channel_mask"] == (
            t.Channels.from_channel_list([network_info.channel]),
        )
    elif channel_mask is not None:
        assert params["channel_mask"] == (network_info.channel_mask,)
    else:
        assert False

    assert params["use_predefined_nwk_panid"] == (True,)
    assert params["nwk_panid"] == (network_info.pan_id,)
    assert params["aps_extended_panid"] == (network_info.extended_pan_id,)
    assert params["nwk_update_id"] == (network_info.nwk_update_id,)
    assert params["network_key"] == (
        zigpy_deconz.api.IndexedKey(index=0, key=network_info.network_key.key),
    )
    assert params["trust_center_address"] == (node_info.ieee,)
    assert params["link_key"] == (
        zigpy_deconz.api.LinkKey(ieee=node_info.ieee, key=network_info.tc_link_key.key),
    )

    if security_level == 0:
        assert params["security_mode"] == (zigpy_deconz.api.SecurityMode.NO_SECURITY,)
    else:
        assert params["security_mode"] == (zigpy_deconz.api.SecurityMode.ONLY_TCLK,)


@patch.object(application, "CHANGE_NETWORK_POLL_TIME", 0.001)
@patch.object(application, "CHANGE_NETWORK_STATE_DELAY", 0.001)
@pytest.mark.parametrize(
    "error, param_overrides, nwk_state_changes, node_state_changes",
    [
        (None, {}, {}, {}),
        (
            None,
            {("aps_designed_coordinator",): 0x00},
            {},
            {"logical_type": zdo_t.LogicalType.Router},
        ),
        (
            None,
            {
                ("aps_extended_panid",): t.EUI64.convert("00:00:00:00:00:00:00:00"),
                ("nwk_extended_panid",): t.EUI64.convert("0D:49:91:99:AE:CD:3C:35"),
            },
            {},
            {},
        ),
        (NetworkNotFormed, {("current_channel",): 0}, {}, {}),
        (
            None,
            {
                ("nwk_frame_counter",): zigpy_deconz.exception.CommandError(
                    "Some error",
                    status=zigpy_deconz.api.Status.UNSUPPORTED,
                    command=None,
                )
            },
            {"network_key.tx_counter": 1073741824},
            {},
        ),
        (
            None,
            {("security_mode",): zigpy_deconz.api.SecurityMode.NO_SECURITY},
            {"security_level": 0},
            {},
        ),
        (
            None,
            {
                (
                    "security_mode",
                ): zigpy_deconz.api.SecurityMode.PRECONFIGURED_NETWORK_KEY
            },
            {"security_level": 5},
            {},
        ),
    ],
)
async def test_load_network_info(
    app,  # noqa: F811
    network_info,
    node_info,
    error,
    param_overrides,
    nwk_state_changes,
    node_state_changes,
):
    """Test that network info is correctly read."""

    params = {
        ("nwk_frame_counter",): network_info.network_key.tx_counter,
        ("aps_designed_coordinator",): 1,
        ("nwk_address",): node_info.nwk,
        ("mac_address",): node_info.ieee,
        ("current_channel",): network_info.channel,
        ("channel_mask",): t.Channels.from_channel_list([network_info.channel]),
        ("use_predefined_nwk_panid",): True,
        ("nwk_panid",): network_info.pan_id,
        ("aps_extended_panid",): network_info.extended_pan_id,
        ("nwk_update_id",): network_info.nwk_update_id,
        ("network_key", 0): zigpy_deconz.api.IndexedKey(
            index=0, key=network_info.network_key.key
        ),
        ("trust_center_address",): node_info.ieee,
        ("link_key", node_info.ieee): zigpy_deconz.api.LinkKey(
            ieee=node_info.ieee, key=network_info.tc_link_key.key
        ),
        ("security_mode",): zigpy_deconz.api.SecurityMode.ONLY_TCLK,
        ("protocol_version",): 0x010E,
    }

    params.update(param_overrides)

    async def read_param(param, *args):
        try:
            value = params[(param.name,) + args]
        except KeyError:
            raise zigpy_deconz.exception.CommandError(
                zigpy_deconz.api.Status.UNSUPPORTED, f"Unsupported: {param!r} {args!r}"
            )

        if isinstance(value, Exception):
            raise value

        return value

    app._api.firmware_version = zigpy_deconz.api.FirmwareVersion(0x26580700)
    app._api.read_parameter = AsyncMock(side_effect=read_param)

    if error is not None:
        with pytest.raises(error):
            await app.load_network_info()

        return

    assert app.state.network_info != network_info
    assert app.state.node_info != node_info

    await app.load_network_info()

    # Almost all of the info matches
    network_info = network_info.replace(
        channel_mask=t.Channels.from_channel_list([network_info.channel]),
        network_key=network_info.network_key.replace(seq=0),
        tc_link_key=network_info.tc_link_key.replace(tx_counter=0),
    )
    merge_objects(network_info, nwk_state_changes)

    assert app.state.network_info == network_info

    assert app.state.node_info == node_info.replace(**node_state_changes)


@patch.object(application, "CHANGE_NETWORK_POLL_TIME", 0.001)
@patch.object(application, "CHANGE_NETWORK_STATE_DELAY", 0.001)
async def test_form_network_fast_without_frame_counter_support(app):  # noqa: F811
    """Test that form_network(fast=True) works when FW doesn't support frame counter."""

    async def write_parameter(param, *args):
        if param == zigpy_deconz.api.NetworkParameter.nwk_frame_counter:
            raise zigpy_deconz.exception.CommandError(
                "Command is unsupported",
                status=zigpy_deconz.api.Status.UNSUPPORTED,
                command=None,
            )

    app._change_network_state = AsyncMock()
    app._api.write_parameter = AsyncMock(side_effect=write_parameter)
    app.backups = AsyncMock()
    app.backups.restore_backup = AsyncMock()

    # This should not raise an error because fast=True sets form_quickly
    await app.form_network(fast=True)

    # Verify that restore_backup was called with create_new=False (due to fast=True)
    assert app.backups.restore_backup.mock_calls == [
        call(backup=ANY, counter_increment=0, allow_incomplete=True, create_new=False)
    ]


@patch.object(application, "CHANGE_NETWORK_POLL_TIME", 0.001)
@patch.object(application, "CHANGE_NETWORK_STATE_DELAY", 0.001)
async def test_reset_network_info_without_frame_counter_support(app):  # noqa: F811
    """Test that reset_network_info works even when FW doesn't support frame counter."""

    async def write_parameter(param, *args):
        if param == zigpy_deconz.api.NetworkParameter.nwk_frame_counter:
            raise zigpy_deconz.exception.CommandError(
                "Command is unsupported",
                status=zigpy_deconz.api.Status.UNSUPPORTED,
                command=None,
            )

    app._change_network_state = AsyncMock()
    app._api.write_parameter = AsyncMock(side_effect=write_parameter)
    app._api.read_parameter = AsyncMock(
        return_value=t.EUI64.convert("00:11:22:33:44:55:66:77")
    )

    # Should not raise an error despite frame counter not being supported
    await app.reset_network_info()

    # Verify that write_parameter was called (including the frame counter attempt)
    assert any(
        mock_call.args[0] == zigpy_deconz.api.NetworkParameter.nwk_frame_counter
        for mock_call in app._api.write_parameter.mock_calls
    )

    # Verify network state changes were awaited
    assert app._change_network_state.mock_calls == [
        call(zigpy_deconz.api.NetworkState.OFFLINE)
    ]



"""Tests for the uart module."""

import logging
from unittest import mock

import pytest
from zigpy.config import (
    CONF_DEVICE_BAUDRATE,
    CONF_DEVICE_FLOW_CONTROL,
    CONF_DEVICE_PATH,
)
import zigpy.serial

from zigpy_deconz import uart


@pytest.fixture
def gw():
    gw = uart.Gateway(mock.MagicMock())
    gw._transport = mock.MagicMock()
    return gw


async def test_connect(monkeypatch):
    api = mock.MagicMock()

    async def mock_conn(loop, protocol_factory, **kwargs):
        protocol = protocol_factory()
        loop.call_soon(protocol.connection_made, None)
        return None, protocol

    monkeypatch.setattr(zigpy.serial, "create_serial_connection", mock_conn)

    await uart.connect(
        {
            CONF_DEVICE_PATH: "/dev/null",
            CONF_DEVICE_BAUDRATE: 115200,
            CONF_DEVICE_FLOW_CONTROL: None,
        },
        api,
    )


def test_send(gw):
    data = b"test"
    gw.send(data)

    packet = b""
    packet += b"\xC0"  # END
    packet += b"test"  # data
    packet += b"\x40\xFE"  # checksum
    packet += b"\xC0"  # END

    gw._transport.write.assert_called_once_with(packet)


def test_close(gw):
    gw.close()
    assert gw._transport.close.call_count == 1


def test_data_received_chunk_frame(gw):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02\x44\xFF\xC0"
    gw.data_received(data[:-4])
    assert gw._api.data_received.call_count == 0
    gw.data_received(data[-4:])
    assert gw._api.data_received.call_count == 1
    assert gw._api.data_received.call_args[0][0] == data[:-3]


def test_data_received_full_frame(gw):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02\x44\xFF\xC0"
    gw.data_received(data)
    assert gw._api.data_received.call_count == 1
    assert gw._api.data_received.call_args[0][0] == data[:-3]


def test_data_received_incomplete_frame(gw):
    data = b"~\x00\x00"
    gw.data_received(data)
    assert gw._api.data_received.call_count == 0


def test_data_received_runt_frame(gw):
    data = b"\x02\x44\xC0"
    gw.data_received(data)
    assert gw._api.data_received.call_count == 0


def test_data_received_extra(gw):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02\x44\xFF\xC0\x00"
    gw.data_received(data)
    assert gw._api.data_received.call_count == 1
    assert gw._api.data_received.call_args[0][0] == data[:-4]
    assert gw._buffer == b"\x00"


def test_data_received_wrong_checksum(gw):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02\x44\xFE\xC0"
    gw.data_received(data)
    assert gw._api.data_received.call_count == 0


def test_data_received_error(gw, caplog):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02\x44\xFF\xC0"

    gw._api.data_received.side_effect = [RuntimeError("error")]

    with caplog.at_level(logging.ERROR):
        gw.data_received(data)

    assert "RuntimeError" in caplog.text and "handling the frame" in caplog.text

    assert gw._api.data_received.call_count == 1
    assert gw._api.data_received.call_args[0][0] == data[:-3]


def test_unescape(gw):
    data = b"\x00\xDB\xDC\x00\xDB\xDD\x00\x00\x00"
    data_unescaped = b"\x00\xC0\x00\xDB\x00\x00\x00"
    r = gw._unescape(data)
    assert r == data_unescaped


def test_unescape_error(gw):
    data = b"\x00\xDB\xDC\x00\xDB\xDD\x00\x00\x00\xDB"
    r = gw._unescape(data)
    assert r is None


def test_escape(gw):
    data = b"\x00\xC0\x00\xDB\x00\x00\x00"
    data_escaped = b"\x00\xDB\xDC\x00\xDB\xDD\x00\x00\x00"
    r = gw._escape(data)
    assert r == data_escaped


def test_checksum(gw):
    data = b"\x07\x01\x00\x08\x00\xaa\x00\x02"
    checksum = b"\x44\xFF"
    r = gw._checksum(data)
    assert r == checksum


def test_connection_lost_exc(gw):
    gw.connection_lost(mock.sentinel.exception)

    conn_lost = gw._api.connection_lost
    assert conn_lost.call_count == 1
    assert conn_lost.call_args[0][0] is mock.sentinel.exception


def test_connection_closed(gw):
    gw.connection_lost(None)

    assert gw._api.connection_lost.call_count == 1



"""Test utils module."""

import asyncio
import logging
from unittest.mock import AsyncMock, patch

import pytest

from zigpy_deconz import utils


async def test_restart_forever(caplog):
    mock = AsyncMock(side_effect=[None, RuntimeError(), RuntimeError(), None])
    func = utils.restart_forever(
        mock,
        restart_delay=0.1,
    )

    with caplog.at_level(logging.DEBUG):
        task = asyncio.create_task(func())
        await asyncio.sleep(0.5)
        task.cancel()

    assert caplog.text.count("failed, restarting...") >= 2
    assert caplog.text.count("RuntimeError") == 2
    assert len(mock.mock_calls) >= 4


@pytest.mark.parametrize(
    ("device_path", "realpath_result", "expected_is_usb"),
    [
        # Platform serial ports (Raspbee)
        ("/dev/conbee", "/dev/ttyS0", False),
        ("/dev/raspbee", "/dev/ttyAMA0", False),
        ("/dev/ttyS0", "/dev/ttyS0", False),
        ("/dev/ttyAMA0", "/dev/ttyAMA0", False),
        ("/dev/ttyS1", "/dev/ttyS1", False),
        ("/dev/ttyAMA1", "/dev/ttyAMA1", False),
        # USB serial ports (Conbee)
        ("/dev/conbee", "/dev/ttyUSB0", True),
        ("/dev/ttyUSB0", "/dev/ttyUSB0", True),
        ("/dev/ttyACM0", "/dev/ttyACM0", True),
        ("/dev/ttyUSB1", "/dev/ttyUSB1", True),
        ("/dev/ttyACM1", "/dev/ttyACM1", True),
        # Symlink to USB serial (Conbee)
        ("/dev/serial/by-id/usb-conbee", "/dev/ttyUSB0", True),
    ],
)
def test_is_usb_serial_port(device_path, realpath_result, expected_is_usb):
    """Test is_usb_serial_port with various device paths and realpath results."""
    with patch("zigpy_deconz.utils.os.path.realpath", return_value=realpath_result):
        result = utils.is_usb_serial_port(device_path)

    assert result == expected_is_usb



"""Tests for zigpy_deconz.types module."""


import pytest
import zigpy.types as zigpy_t

import zigpy_deconz.types as t


def test_deconz_address_group():
    data = b"\x01\x55\xaa"
    extra = b"the rest of the owl"

    addr, rest = t.DeconzAddress.deserialize(data + extra)
    assert rest == extra
    assert addr.address_mode == t.AddressMode.GROUP
    assert addr.address_mode == 1
    assert addr.address == 0xAA55

    assert addr.serialize() == data

    zigpy_addr = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.Group, address=0xAA55
    )
    assert addr.as_zigpy_type() == zigpy_addr

    converted_addr = t.DeconzAddress.from_zigpy_type(zigpy_addr)
    assert converted_addr == addr


def test_deconz_address_nwk():
    data = b"\x02\x55\xaa"
    extra = b"the rest of the owl"

    addr, rest = t.DeconzAddress.deserialize(data + extra)
    assert rest == extra
    assert addr.address_mode == t.AddressMode.NWK
    assert addr.address_mode == 2
    assert addr.address == 0xAA55

    assert addr.serialize() == data

    zigpy_addr = zigpy_t.AddrModeAddress(addr_mode=zigpy_t.AddrMode.NWK, address=0xAA55)
    assert addr.as_zigpy_type() == zigpy_addr

    converted_addr = t.DeconzAddress.from_zigpy_type(zigpy_addr)
    assert converted_addr == addr


def test_deconz_address_nwk_broadcast():
    data = b"\x02\xfc\xff"
    extra = b"the rest of the owl"

    addr, rest = t.DeconzAddress.deserialize(data + extra)
    assert rest == extra
    assert addr.address_mode == t.AddressMode.NWK
    assert addr.address_mode == 2
    assert addr.address == 0xFFFC

    assert addr.serialize() == data

    zigpy_addr = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.Broadcast,
        address=zigpy_t.BroadcastAddress.ALL_ROUTERS_AND_COORDINATOR,
    )
    assert addr.as_zigpy_type() == zigpy_addr

    converted_addr = t.DeconzAddress.from_zigpy_type(zigpy_addr)
    assert converted_addr == addr


def test_deconz_address_ieee():
    data = b"\x03\x55\xaa\xbb\xcc\xdd\xee\xef\xbe"
    extra = b"the rest of the owl"

    addr, rest = t.DeconzAddress.deserialize(data + extra)
    assert rest == extra
    assert addr.address_mode == t.AddressMode.IEEE
    assert addr.address_mode == 3
    assert addr.address[0] == 0x55
    assert addr.address[1] == 0xAA
    assert addr.address[2] == 0xBB
    assert addr.address[3] == 0xCC
    assert addr.address[4] == 0xDD
    assert addr.address[5] == 0xEE
    assert addr.address[6] == 0xEF
    assert addr.address[7] == 0xBE

    assert addr.serialize() == data

    zigpy_addr = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.IEEE,
        address=zigpy_t.EUI64.convert("BE:EF:EE:DD:CC:BB:AA:55"),
    )
    assert addr.as_zigpy_type() == zigpy_addr

    converted_addr = t.DeconzAddress.from_zigpy_type(zigpy_addr)
    assert converted_addr == addr


def test_deconz_address_nwk_and_ieee():
    data = b"\x04\x55\xaa\x88\x99\xbb\xcc\xdd\xee\xef\xbe"
    extra = b"the rest of the owl"

    addr, rest = t.DeconzAddress.deserialize(data + extra)
    assert rest == extra
    assert addr.address_mode == t.AddressMode.NWK_AND_IEEE
    assert addr.address_mode == 4
    assert addr.ieee[0] == 0x88
    assert addr.ieee[1] == 0x99
    assert addr.ieee[2] == 0xBB
    assert addr.ieee[3] == 0xCC
    assert addr.ieee[4] == 0xDD
    assert addr.ieee[5] == 0xEE
    assert addr.ieee[6] == 0xEF
    assert addr.ieee[7] == 0xBE
    assert addr.address == 0xAA55

    assert addr.serialize() == data

    zigpy_addr = zigpy_t.AddrModeAddress(
        addr_mode=zigpy_t.AddrMode.IEEE,
        address=zigpy_t.EUI64.convert("BE:EF:EE:DD:CC:BB:99:88"),
    )
    assert addr.as_zigpy_type() == zigpy_addr


def test_bytes():
    data = b"abcde\x00\xff"

    r, rest = t.Bytes.deserialize(data)
    assert rest == b""
    assert r == data

    assert r.serialize() == data


def test_addr_ep_nwk():
    data = b"\x02\xaa\x55\xcc"
    extra = b"\x00extra data\xff"

    r, rest = t.DeconzAddressEndpoint.deserialize(data + extra)
    assert rest == extra
    assert r.address_mode == t.AddressMode.NWK
    assert r.address == 0x55AA
    assert r.endpoint == 0xCC


def test_addr_ep_ieee():
    data = b"\x0387654321\xcc"
    extra = b"\x00extra data\xff"

    r, rest = t.DeconzAddressEndpoint.deserialize(data + extra)
    assert rest == extra
    assert r.address_mode == t.AddressMode.IEEE
    assert repr(r.address) == "31:32:33:34:35:36:37:38"
    assert r.endpoint == 0xCC


def test_deconz_addr_ep():
    data = b"\x01\xaa\x55"
    extra = b"the rest of the owl"

    r, rest = t.DeconzAddressEndpoint.deserialize(data + extra)
    assert rest == extra
    assert r.address_mode == t.AddressMode.GROUP
    assert r.address == 0x55AA
    assert r.serialize() == data
    a = t.DeconzAddressEndpoint()
    a.address_mode = 1
    a.address = 0x55AA
    assert a.serialize() == data

    data = b"\x02\xaa\x55\xcc"
    r, rest = t.DeconzAddressEndpoint.deserialize(data + extra)
    assert rest == extra
    assert r.address_mode == t.AddressMode.NWK
    assert r.address == 0x55AA
    assert r.endpoint == 0xCC
    assert r.serialize() == data
    a = t.DeconzAddressEndpoint()
    a.address_mode = 2
    a.address = 0x55AA
    with pytest.raises(TypeError):
        a.serialize()
    a.endpoint = 0xCC
    assert a.serialize() == data

    data = b"\x03\x31\x32\x33\x34\x35\x36\x37\x38\xcc"
    r, rest = t.DeconzAddressEndpoint.deserialize(data + extra)
    assert rest == extra
    assert r.address_mode == t.AddressMode.IEEE
    assert r.address == [0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38]
    assert r.endpoint == 0xCC
    assert r.serialize() == data
    a = t.DeconzAddressEndpoint()
    a.address_mode = 3
    a.address = [0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38]
    with pytest.raises(TypeError):
        a.serialize()
    a.endpoint = 0xCC
    assert a.serialize() == data


def test_nwklist():
    assert t.NWKList([]).serialize() == b"\x00"
    assert t.NWKList([0x1234]).serialize() == b"\x01" + t.NWK(0x1234).serialize()
    assert (
        t.NWKList([0x1234, 0x5678]).serialize()
        == b"\x02" + t.NWK(0x1234).serialize() + t.NWK(0x5678).serialize()
    )

    assert t.NWKList.deserialize(b"\x00abc") == (t.NWKList([]), b"abc")
    assert t.NWKList.deserialize(b"\x01\x34\x12abc") == (t.NWKList([0x1234]), b"abc")
    assert t.NWKList.deserialize(b"\x02\x34\x12\x78\x56abc") == (
        t.NWKList([0x1234, 0x5678]),
        b"abc",
    )


def test_serialize_dict():
    assert (
        t.serialize_dict(
            {"foo": 1, "bar": 2, "baz": None},
            {"foo": t.uint8_t, "bar": t.uint16_t, "baz": t.uint8_t},
        )
        == b"\x01\x02\x00"
    )



"""Test api module."""

import asyncio
import collections
import inspect
import logging
import sys

import pytest
import zigpy.config
import zigpy.types as zigpy_t

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
else:
    from asyncio import timeout as asyncio_timeout

from zigpy_deconz import api as deconz_api, types as t, uart
import zigpy_deconz.exception
import zigpy_deconz.zigbee.application

from .async_mock import AsyncMock, MagicMock, call, patch

DEVICE_CONFIG = {zigpy.config.CONF_DEVICE_PATH: "/dev/null"}


@pytest.fixture
async def gateway():
    return uart.Gateway(api=None)


@pytest.fixture
async def api(gateway, mock_command_rsp):
    loop = asyncio.get_running_loop()

    async def mock_connect(config, api):
        transport = MagicMock()
        transport.close = MagicMock(
            side_effect=lambda: loop.call_soon(gateway.connection_lost, None)
        )

        gateway._api = api
        gateway.connection_made(transport)

        return gateway

    with patch("zigpy_deconz.uart.connect", side_effect=mock_connect):
        controller = MagicMock(
            spec_set=zigpy_deconz.zigbee.application.ControllerApplication
        )
        api = deconz_api.Deconz(
            controller, {zigpy.config.CONF_DEVICE_PATH: "/dev/null"}
        )

        mock_command_rsp(
            command_id=deconz_api.CommandId.device_state,
            params={},
            rsp={
                "status": deconz_api.Status.SUCCESS,
                "frame_length": t.uint16_t(8),
                "device_state": deconz_api.DeviceState(
                    network_state=deconz_api.NetworkState2.CONNECTED,
                    device_state=(
                        deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                    ),
                ),
                "reserved1": t.uint8_t(0),
                "reserved2": t.uint8_t(0),
            },
        )

        mock_command_rsp(
            command_id=deconz_api.CommandId.read_parameter,
            params={
                "parameter_id": deconz_api.NetworkParameter.protocol_version,
                "parameter": t.Bytes(b""),
            },
            rsp={
                "status": deconz_api.Status.SUCCESS,
                "frame_length": t.uint16_t(10),
                "payload_length": t.uint16_t(3),
                "parameter_id": deconz_api.NetworkParameter.protocol_version,
                "parameter": t.Bytes(t.uint16_t(270).serialize()),
            },
        )

        mock_command_rsp(
            command_id=deconz_api.CommandId.version,
            params={"reserved": t.uint8_t(0)},
            rsp={
                "status": deconz_api.Status.SUCCESS,
                "frame_length": t.uint16_t(9),
                "version": deconz_api.FirmwareVersion(645400320),
            },
        )

        yield api


@pytest.fixture
async def mock_command_rsp(gateway):
    def inner(command_id, params, rsp, *, rsp_command=None, replace=False):
        if (
            getattr(getattr(gateway.send, "side_effect", None), "_handlers", None)
            is None
        ):

            def receiver(data):
                command, _ = deconz_api.Command.deserialize(data)
                tx_schema, _ = deconz_api.COMMAND_SCHEMAS[command.command_id]
                schema = {}

                for k, v in tx_schema.items():
                    if v in (deconz_api.FRAME_LENGTH, deconz_api.PAYLOAD_LENGTH):
                        v = t.uint16_t
                    elif not inspect.isclass(v):
                        v = type(v)

                    schema[k] = v

                kwargs, rest = t.deserialize_dict(command.payload, schema)

                for params, rsp_command, mock in receiver._handlers[command.command_id]:
                    if rsp_command is None:
                        rsp_command = command.command_id

                    if all(kwargs[k] == v for k, v in params.items()):
                        _, rx_schema = deconz_api.COMMAND_SCHEMAS[rsp_command]
                        ret = mock(**kwargs)

                        asyncio.get_running_loop().call_soon(
                            gateway._api.data_received,
                            deconz_api.Command(
                                command_id=rsp_command,
                                seq=command.seq,
                                payload=t.serialize_dict(ret, rx_schema),
                            ).serialize(),
                        )

            receiver._handlers = collections.defaultdict(list)
            gateway.send = MagicMock(side_effect=receiver)

        if replace:
            gateway.send.side_effect._handlers[command_id].clear()

        mock = MagicMock(return_value=rsp)
        gateway.send.side_effect._handlers[command_id].append(
            (params, rsp_command, mock)
        )

        return mock

    return inner


def send_network_state(
    api,
    network_state: deconz_api.NetworkState2 = deconz_api.NetworkState2.CONNECTED,
    device_state: deconz_api.DeviceStateFlags = (
        deconz_api.DeviceStateFlags.APSDE_DATA_CONFIRM
    ),
):
    _, rx_schema = deconz_api.COMMAND_SCHEMAS[deconz_api.CommandId.device_state_changed]

    data = deconz_api.Command(
        command_id=deconz_api.CommandId.device_state_changed,
        seq=api._seq,
        payload=t.serialize_dict(
            {
                "status": deconz_api.Status.SUCCESS,
                "frame_length": t.uint16_t(7),
                "device_state": deconz_api.DeviceState(
                    network_state=network_state,
                    device_state=device_state,
                ),
                "reserved": t.uint8_t(0),
            },
            rx_schema,
        ),
    ).serialize()

    asyncio.get_running_loop().call_later(0.01, api.data_received, data)


async def test_connect(api, mock_command_rsp):
    await api.connect()


async def test_connect_failure(api, mock_command_rsp):
    transport = None

    def mock_version(*args, **kwargs):
        nonlocal transport
        transport = api._uart._transport

        raise asyncio.TimeoutError()

    with patch.object(api, "version", side_effect=mock_version):
        # We connect but fail to probe
        with pytest.raises(asyncio.TimeoutError):
            await api.connect()

    assert api._uart is None
    assert len(transport.close.mock_calls) == 1


async def test_close(api):
    await api.connect()

    uart = api._uart
    uart.disconnect = AsyncMock()

    await api.disconnect()
    assert api._uart is None
    assert uart.disconnect.call_count == 1


def test_commands():
    for cmd, (tx_schema, rx_schema) in deconz_api.COMMAND_SCHEMAS.items():
        assert isinstance(cmd, deconz_api.CommandId)
        assert isinstance(tx_schema, dict) or tx_schema is None
        assert isinstance(rx_schema, dict)


async def test_command(api):
    await api.connect()

    addr = t.DeconzAddress()
    addr.address_mode = t.AddressMode.NWK
    addr.address = t.NWK(0x0000)

    params = {
        "status": deconz_api.Status.SUCCESS,
        "frame_length": t.uint16_t(61),
        "payload_length": t.uint16_t(54),
        "device_state": deconz_api.DeviceState(
            network_state=deconz_api.NetworkState2.CONNECTED,
            device_state=(
                deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
            ),
        ),
        "dst_addr": addr,
        "dst_ep": t.uint8_t(0),
        "src_addr": addr,
        "src_ep": t.uint8_t(0),
        "profile_id": t.uint16_t(0),
        "cluster_id": t.uint16_t(32772),
        "asdu": t.LongOctetString(
            b"\x0f\x00\x00\x00\x1a\x01\x04\x01\x00\x04\x00\x05\x00\x00\x06"
            b"\x00\n\x00\x19\x00\x01\x05\x04\x01\x00 \x00\x00\x05\x02\x05"
        ),
        "reserved1": t.uint8_t(0),
        "reserved2": t.uint8_t(175),
        "lqi": t.uint8_t(69),
        "reserved3": t.uint8_t(189),
        "reserved4": t.uint8_t(82),
        "reserved5": t.uint8_t(0),
        "reserved6": t.uint8_t(0),
        "rssi": t.int8s(27),
    }

    data = deconz_api.Command(
        command_id=deconz_api.CommandId.aps_data_indication,
        seq=api._seq,
        payload=t.serialize_dict(
            params,
            deconz_api.COMMAND_SCHEMAS[deconz_api.CommandId.aps_data_indication][1],
        ),
    ).serialize()

    asyncio.get_running_loop().call_later(0.01, api.data_received, data)

    rsp = await api._command(
        cmd=deconz_api.CommandId.aps_data_indication,
        flags=t.DataIndicationFlags.Include_Both_NWK_And_IEEE,
    )
    assert rsp == params


async def test_command_lock(api, mock_command_rsp):
    await api.connect()

    for i in range(4):
        mock_command_rsp(
            command_id=deconz_api.CommandId.version,
            params={"reserved": t.uint8_t(i)},
            rsp={
                "status": deconz_api.Status.SUCCESS,
                "frame_length": t.uint16_t(9),
                "version": deconz_api.FirmwareVersion(i),
            },
            replace=(i == 0),
        )

    async with api._command_lock:
        tasks = [
            asyncio.create_task(
                api._command(cmd=deconz_api.CommandId.version, reserved=0)
            ),
            asyncio.create_task(
                api._command(cmd=deconz_api.CommandId.version, reserved=1)
            ),
            asyncio.create_task(
                api._command(cmd=deconz_api.CommandId.version, reserved=2)
            ),
            asyncio.create_task(
                api._command(cmd=deconz_api.CommandId.version, reserved=3)
            ),
        ]

        await asyncio.sleep(0.1)
        assert not any(t.done() for t in tasks)

    responses = await asyncio.gather(*tasks)

    for index, rsp in enumerate(responses):
        assert rsp["version"] == index


async def test_command_timeout(api):
    await api.connect()

    with patch.object(deconz_api, "COMMAND_TIMEOUT", 0.1):
        with pytest.raises(asyncio.TimeoutError):
            await api._command(
                cmd=deconz_api.CommandId.change_network_state,
                network_state=deconz_api.NetworkState.OFFLINE,
            )


async def test_command_not_connected(api):
    api._uart = None

    with pytest.raises(deconz_api.CommandError):
        await api._command(cmd=deconz_api.CommandId.version, reserved=0)


async def test_data_received(api, mock_command_rsp):
    await api.connect()

    src_addr = t.DeconzAddress()
    src_addr.address_mode = t.AddressMode.NWK
    src_addr.address = t.NWK(0xE695)

    dst_addr = t.DeconzAddress()
    dst_addr.address_mode = t.AddressMode.NWK
    dst_addr.address = t.NWK(0x0000)

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_indication,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(80),
            "payload_length": t.uint16_t(73),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                ),
            ),
            "dst_addr": dst_addr,
            "dst_ep": t.uint8_t(1),
            "src_addr": src_addr,
            "src_ep": t.uint8_t(1),
            "profile_id": t.uint16_t(260),
            "cluster_id": t.uint16_t(0x0000),
            "asdu": t.LongOctetString(
                b"\x18\x1b\x01\x04\x00\x00B\x0eIKEA of Sweden"
                b"\x05\x00\x00B\x17TRADFRI wireless dimmer"
            ),
            "reserved1": t.uint8_t(0),
            "reserved2": t.uint8_t(175),
            "lqi": t.uint8_t(255),
            "reserved3": t.uint8_t(142),
            "reserved4": t.uint8_t(98),
            "reserved5": t.uint8_t(0),
            "reserved6": t.uint8_t(0),
            "rssi": t.int8s(-49),
        },
    )

    # Unsolicited device_state_changed
    api.data_received(bytes.fromhex("0e2f000700ae00"))

    await asyncio.sleep(0.1)

    api._app.packet_received.assert_called_once_with(
        zigpy_t.ZigbeePacket(
            src=zigpy_t.AddrModeAddress(addr_mode=zigpy_t.AddrMode.NWK, address=0xE695),
            src_ep=1,
            dst=zigpy_t.AddrModeAddress(addr_mode=zigpy_t.AddrMode.NWK, address=0x0000),
            dst_ep=1,
            source_route=None,
            extended_timeout=False,
            tsn=None,
            profile_id=260,
            cluster_id=0x0000,
            data=zigpy_t.SerializableBytes(
                b"\x18\x1b\x01\x04\x00\x00B\x0eIKEA of Sweden"
                b"\x05\x00\x00B\x17TRADFRI wireless dimmer"
            ),
            tx_options=zigpy_t.TransmitOptions.NONE,
            radius=0,
            non_member_radius=0,
            lqi=255,
            rssi=-49,
        )
    )


async def test_read_parameter(api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.read_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.nwk_update_id,
            "parameter": t.Bytes(b""),
        },
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "payload_length": t.uint16_t(2),
            "parameter_id": deconz_api.NetworkParameter.nwk_update_id,
            "parameter": t.Bytes(b"\x00"),
        },
    )

    mock_command_rsp(
        command_id=deconz_api.CommandId.read_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.network_key,
            "parameter": t.Bytes(b"\x00"),
        },
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(25),
            "payload_length": t.uint16_t(18),
            "parameter_id": deconz_api.NetworkParameter.network_key,
            "parameter": t.Bytes(b"\x00M\x07p\xb6\x0b|\x90\xad\\\x07\x8a8\xa9M\xf6["),
        },
    )

    rsp = await api.read_parameter(deconz_api.NetworkParameter.nwk_update_id)
    assert rsp == 0x00

    rsp = await api.read_parameter(deconz_api.NetworkParameter.network_key, 0)
    assert rsp == deconz_api.IndexedKey(
        index=0,
        key=deconz_api.KeyData.convert(
            "4d:07:70:b6:0b:7c:90:ad:5c:07:8a:38:a9:4d:f6:5b"
        ),
    )


async def test_write_parameter(api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.write_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.watchdog_ttl,
            "parameter": t.uint32_t(600).serialize(),
        },
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(8),
            "payload_length": t.uint16_t(1),
            "parameter_id": deconz_api.NetworkParameter.watchdog_ttl,
        },
    )

    await api.write_parameter(deconz_api.NetworkParameter.watchdog_ttl, 600)


async def test_write_parameter_failure(api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.write_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.watchdog_ttl,
            "parameter": t.uint32_t(600).serialize(),
        },
        rsp={
            "status": deconz_api.Status.INVALID_VALUE,
            "frame_length": t.uint16_t(8),
            "payload_length": t.uint16_t(1),
            "parameter_id": deconz_api.NetworkParameter.watchdog_ttl,
        },
    )

    with pytest.raises(deconz_api.CommandError):
        await api.write_parameter(deconz_api.NetworkParameter.watchdog_ttl, 600)


@pytest.mark.parametrize(
    "protocol_ver, firmware_ver",
    [
        (0x010A, 0x123405DD),
        (0x010B, 0x123405DD),
        (0x010A, 0x123407DD),
        (0x010B, 0x123407DD),
    ],
)
async def test_version(protocol_ver, firmware_ver, api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.read_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.protocol_version,
            "parameter": t.Bytes(b""),
        },
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(10),
            "payload_length": t.uint16_t(3),
            "parameter_id": deconz_api.NetworkParameter.protocol_version,
            "parameter": t.Bytes(t.uint16_t(protocol_ver).serialize()),
        },
        replace=True,
    )

    mock_command_rsp(
        command_id=deconz_api.CommandId.version,
        params={"reserved": t.uint8_t(0)},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "version": deconz_api.FirmwareVersion(firmware_ver),
        },
        replace=True,
    )

    r = await api.version()
    assert r == firmware_ver

    assert api.protocol_version == protocol_ver
    assert api.firmware_version == firmware_ver


@pytest.mark.parametrize(
    "data, network_state",
    ((0x00, "OFFLINE"), (0x01, "JOINING"), (0x02, "CONNECTED"), (0x03, "LEAVING")),
)
def test_device_state_network_state(data, network_state):
    """Test device state flag."""
    extra = b"the rest of the data\xaa\x55"

    for other_fields in (0x04, 0x08, 0x0C, 0x10, 0x24, 0x28, 0x30, 0x2C):
        new_data = t.uint8_t(data | other_fields).serialize()
        state, rest = deconz_api.DeviceState.deserialize(new_data + extra)
        assert rest == extra
        assert state.network_state == deconz_api.NetworkState[network_state]
        assert state.serialize() == new_data


@pytest.mark.parametrize(
    "value, name",
    (
        (0x00, "SUCCESS"),
        (0xA0, "APS_ASDU_TOO_LONG"),
        (0x01, "MAC_PAN_AT_CAPACITY"),
        (0xC9, "NWK_UNSUPPORTED_ATTRIBUTE"),
        (0xFE, "undefined_0xfe"),
    ),
)
def test_tx_status(value, name):
    """Test tx status undefined values."""
    i = deconz_api.TXStatus(value)
    assert i == value
    assert i.value == value
    assert i.name == name

    extra = b"\xaa\55"
    data = t.uint8_t(value).serialize()
    status, rest = deconz_api.TXStatus.deserialize(data + extra)
    assert rest == extra
    assert isinstance(status, deconz_api.TXStatus)
    assert status == value
    assert status.value == value
    assert status.name == name


@pytest.mark.parametrize("relays", (None, [], [0x1234, 0x5678]))
async def test_aps_data_request_relays(relays, api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_request,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "payload_length": t.uint16_t(2),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                ),
            ),
            "request_id": t.uint8_t(0x00),
        },
    )

    await api.aps_data_request(
        req_id=0x00,
        dst_addr_ep=t.DeconzAddressEndpoint.deserialize(b"\x02\xaa\x55\x01")[0],
        profile=0x0104,
        cluster=0x0007,
        src_ep=0x01,
        aps_payload=b"aps payload",
        relays=relays,
    )

    with pytest.raises(ValueError) as exc:
        await api.aps_data_request(
            req_id=0x00,
            dst_addr_ep=t.DeconzAddressEndpoint.deserialize(b"\x02\xaa\x55\x01")[0],
            profile=0x0104,
            cluster=0x0007,
            src_ep=None,  # This is not possible
            aps_payload=b"aps payload",
        )

        assert "has non-trailing optional argument" in str(exc.value)


async def test_aps_data_request_retries_failure(api, mock_command_rsp):
    await api.connect()

    mock_rsp = mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_request,
        params={},
        rsp={
            "status": deconz_api.Status.FAILURE,
            "frame_length": t.uint16_t(9),
            "payload_length": t.uint16_t(2),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                ),
            ),
            "request_id": t.uint8_t(0x00),
        },
    )

    with pytest.raises(deconz_api.CommandError):
        await api.aps_data_request(
            req_id=0x00,
            dst_addr_ep=t.DeconzAddressEndpoint.deserialize(b"\x02\xaa\x55\x01")[0],
            profile=0x0104,
            cluster=0x0007,
            src_ep=1,
            aps_payload=b"aps payload",
        )

    assert len(mock_rsp.mock_calls) == 1


async def test_aps_data_request_locking(caplog, api, mock_command_rsp):
    await api.connect()

    # No free slots
    send_network_state(api, device_state=deconz_api.DeviceStateFlags.APSDE_DATA_CONFIRM)

    await asyncio.sleep(0.1)

    mock_rsp = mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_request,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "payload_length": t.uint16_t(2),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                ),
            ),
            "request_id": t.uint8_t(0x00),
        },
    )

    with caplog.at_level(logging.DEBUG):
        send = asyncio.create_task(
            api.aps_data_request(
                req_id=0x00,
                dst_addr_ep=t.DeconzAddressEndpoint.deserialize(b"\x02\xaa\x55\x01")[0],
                profile=0x0104,
                cluster=0x0007,
                src_ep=1,
                aps_payload=b"aps payload",
            )
        )

        await asyncio.sleep(0.1)

    assert "Waiting for free slots to become available" in caplog.text

    assert len(mock_rsp.mock_calls) == 0

    send_network_state(
        api,
        device_state=deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE,
    )

    await send

    assert len(mock_rsp.mock_calls) == 1


async def test_connection_lost(api):
    await api.connect()

    app = api._app = MagicMock()

    err = RuntimeError()
    api.connection_lost(err)

    app.connection_lost.assert_called_once_with(err)


async def test_unknown_command(api, caplog):
    await api.connect()

    assert 0xFF not in deconz_api.COMMAND_SCHEMAS

    with caplog.at_level(logging.WARNING):
        api.data_received(b"\xFF\xAA\xBB")

    assert (
        "Unknown command received: Command(command_id=<CommandId.undefined_0xff: 255>,"
        " seq=170, payload=b'\\xbb')"
    ) in caplog.text


async def test_bad_command_parsing(api, caplog):
    await api.connect()

    assert 0xFF not in deconz_api.COMMAND_SCHEMAS

    with caplog.at_level(logging.DEBUG):
        api.data_received(
            bytes.fromhex(
                "172c002f0028002e02000000020000000000"
                "028011000300000010400f3511472b004000"
                # "2b000000af45838600001b"  # truncated
            )
        )

    assert (
        "Failed to parse command Command(command_id="
        "<CommandId.aps_data_indication: 23>"
    ) in caplog.text

    caplog.clear()

    with caplog.at_level(logging.DEBUG):
        api.data_received(bytes.fromhex("0d03000d0000077826") + b"TEST")

    assert (
        "Unparsed data remains after frame" in caplog.text and "b'TEST'" in caplog.text
    )


async def test_bad_response_status(api, mock_command_rsp):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.write_parameter,
        params={
            "parameter_id": deconz_api.NetworkParameter.nwk_update_id,
            "parameter": t.uint8_t(123).serialize(),
        },
        rsp={
            "status": deconz_api.Status.FAILURE,
            "frame_length": t.uint16_t(8),
            "payload_length": t.uint16_t(1),
            "parameter_id": deconz_api.NetworkParameter.nwk_update_id,
        },
    )

    with pytest.raises(deconz_api.CommandError) as exc:
        await api.write_parameter(deconz_api.NetworkParameter.nwk_update_id, 123)

    assert isinstance(exc.value, deconz_api.CommandError)
    assert exc.value.status == deconz_api.Status.FAILURE


async def test_data_poller(api, mock_command_rsp):
    await api.connect()

    dst_addr_ep = t.DeconzAddressEndpoint()
    dst_addr_ep.address_mode = t.AddressMode.NWK
    dst_addr_ep.address = t.NWK(0x0000)
    dst_addr_ep.endpoint = t.uint8_t(0)

    src_addr = t.DeconzAddress()
    src_addr.address_mode = t.AddressMode.NWK
    src_addr.address = t.NWK(0xE695)

    dst_addr = t.DeconzAddress()
    dst_addr.address_mode = t.AddressMode.NWK
    dst_addr.address = t.NWK(0x0000)

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_confirm,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(19),
            "payload_length": t.uint16_t(12),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    # Include a data indication flag to trigger a poll
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                    | deconz_api.DeviceStateFlags.APSDE_DATA_INDICATION
                ),
            ),
            "request_id": t.uint8_t(16),
            "dst_addr": dst_addr_ep,
            "src_ep": t.uint8_t(0),
            "confirm_status": deconz_api.TXStatus.SUCCESS,
            "reserved1": t.uint8_t(0),
            "reserved2": t.uint8_t(0),
            "reserved3": t.uint8_t(0),
            "reserved4": t.uint8_t(0),
        },
    )

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_indication,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(80),
            "payload_length": t.uint16_t(73),
            "device_state": deconz_api.DeviceState(
                network_state=deconz_api.NetworkState2.CONNECTED,
                device_state=(
                    deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                ),
            ),
            "dst_addr": dst_addr,
            "dst_ep": t.uint8_t(1),
            "src_addr": src_addr,
            "src_ep": t.uint8_t(1),
            "profile_id": t.uint16_t(260),
            "cluster_id": t.uint16_t(0x0000),
            "asdu": t.LongOctetString(
                b"\x18\x1b\x01\x04\x00\x00B\x0eIKEA of Sweden"
                b"\x05\x00\x00B\x17TRADFRI wireless dimmer"
            ),
            "reserved1": t.uint8_t(0),
            "reserved2": t.uint8_t(175),
            "lqi": t.uint8_t(255),
            "reserved3": t.uint8_t(142),
            "reserved4": t.uint8_t(98),
            "reserved5": t.uint8_t(0),
            "reserved6": t.uint8_t(0),
            "rssi": t.int8s(-49),
        },
    )

    # Take us offline for a moment
    send_network_state(api, network_state=deconz_api.NetworkState2.OFFLINE)
    await asyncio.sleep(0.1)

    # Bring us back online with just a data confirmation to kick things off
    send_network_state(
        api,
        network_state=deconz_api.NetworkState2.CONNECTED,
        device_state=deconz_api.DeviceStateFlags.APSDE_DATA_CONFIRM,
    )

    await asyncio.sleep(0.1)

    # Both callbacks have been called
    api._app.handle_tx_confirm.assert_called_once_with(16, deconz_api.TXStatus.SUCCESS)
    assert len(api._app.packet_received.mock_calls) == 1

    # The task is cancelled on close
    task = api._data_poller_task
    await api.disconnect()
    assert api._data_poller_task is None
    assert task.done()


async def test_get_device_state(api, mock_command_rsp):
    await api.connect()

    device_state = deconz_api.DeviceState(
        network_state=deconz_api.NetworkState2.CONNECTED,
        device_state=(
            deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
        ),
    )

    mock_command_rsp(
        command_id=deconz_api.CommandId.device_state,
        params={},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(8),
            "device_state": device_state,
            "reserved1": t.uint8_t(0),
            "reserved2": t.uint8_t(0),
        },
    )

    assert (await api.get_device_state()) == device_state


async def test_change_network_state(api, mock_command_rsp):
    api._command = AsyncMock()
    await api.change_network_state(new_state=deconz_api.NetworkState.OFFLINE)

    assert api._command.mock_calls == [
        call(
            deconz_api.CommandId.change_network_state,
            network_state=deconz_api.NetworkState.OFFLINE,
        )
    ]


async def test_add_neighbour(api, mock_command_rsp):
    api._command = AsyncMock()
    await api.add_neighbour(
        nwk=0x1234,
        ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
        mac_capability_flags=0x12,
    )

    assert api._command.mock_calls == [
        call(
            deconz_api.CommandId.update_neighbor,
            action=deconz_api.UpdateNeighborAction.ADD,
            nwk=0x1234,
            ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
            mac_capability_flags=0x12,
        )
    ]


async def test_add_neighbour_conbee3_success(api):
    api._command = AsyncMock(wraps=api._command)
    api._uart = AsyncMock()

    # Simulate a good but invalid response from the Conbee III
    asyncio.get_running_loop().call_later(
        0.001,
        lambda: api.data_received(
            b"\x1d" + bytes([api._seq - 1]) + b"\x00\x06\x00\x01"
        ),
    )

    await api.add_neighbour(
        nwk=0x1234,
        ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
        mac_capability_flags=0x12,
    )

    assert api._command.mock_calls == [
        call(
            deconz_api.CommandId.update_neighbor,
            action=deconz_api.UpdateNeighborAction.ADD,
            nwk=0x1234,
            ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
            mac_capability_flags=0x12,
        )
    ]


async def test_add_neighbour_conbee3_failure(api):
    api._command = AsyncMock(wraps=api._command)
    api._uart = AsyncMock()

    # Simulate a bad response from the Conbee III
    asyncio.get_running_loop().call_later(
        0.001,
        lambda: api.data_received(
            b"\x1d" + bytes([api._seq - 1]) + b"\x01\x06\x00\x01"
        ),
    )

    with pytest.raises(deconz_api.CommandError):
        await api.add_neighbour(
            nwk=0x1234,
            ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
            mac_capability_flags=0x12,
        )

    assert api._command.mock_calls == [
        call(
            deconz_api.CommandId.update_neighbor,
            action=deconz_api.UpdateNeighborAction.ADD,
            nwk=0x1234,
            ieee=t.EUI64.convert("aa:bb:cc:dd:11:22:33:44"),
            mac_capability_flags=0x12,
        )
    ]


async def test_cb3_device_state_callback_bug(api, mock_command_rsp):
    mock_command_rsp(
        command_id=deconz_api.CommandId.version,
        params={"reserved": t.uint8_t(0)},
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "version": deconz_api.FirmwareVersion(0x26450900),
        },
        replace=True,
    )

    await api.connect()

    device_state = deconz_api.DeviceState(
        network_state=deconz_api.NetworkState2.CONNECTED,
        device_state=deconz_api.DeviceStateFlags.APSDE_DATA_CONFIRM,
    )

    assert api._device_state != device_state

    _, rx_schema = deconz_api.COMMAND_SCHEMAS[deconz_api.CommandId.device_state]
    api.data_received(
        deconz_api.Command(
            command_id=deconz_api.CommandId.device_state,
            seq=api._seq,
            payload=t.serialize_dict(
                {
                    "status": deconz_api.Status.SUCCESS,
                    "frame_length": t.uint16_t(8),
                    "device_state": device_state,
                    "reserved1": t.uint8_t(0),
                    "reserved2": t.uint8_t(0),
                },
                rx_schema,
            ),
        ).serialize()
    )

    await asyncio.sleep(0.01)

    assert api._device_state == device_state


async def test_firmware_responding_with_wrong_type_with_correct_seq(
    api, mock_command_rsp, caplog
):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_confirm,
        params={},
        # Completely different response
        rsp_command=deconz_api.CommandId.version,
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "version": deconz_api.FirmwareVersion(0x26450900),
        },
    )

    with caplog.at_level(logging.DEBUG):
        with pytest.raises(asyncio.TimeoutError):
            # We wait beyond 500ms to make sure it triggers
            async with asyncio_timeout(0.6):
                await api.send_command(deconz_api.CommandId.aps_data_confirm)

    assert (
        "Firmware responded incorrectly (Response is mismatched! Sent"
        " <CommandId.aps_data_confirm: 4>, received <CommandId.version: 13>), retrying"
    ) in caplog.text


async def test_firmware_responding_with_wrong_type_with_correct_seq_eventual_response(
    api, mock_command_rsp, caplog
):
    await api.connect()

    mock_command_rsp(
        command_id=deconz_api.CommandId.aps_data_confirm,
        params={},
        # Completely different response
        rsp_command=deconz_api.CommandId.version,
        rsp={
            "status": deconz_api.Status.SUCCESS,
            "frame_length": t.uint16_t(9),
            "version": deconz_api.FirmwareVersion(0x26450900),
        },
    )

    with caplog.at_level(logging.DEBUG):
        _, rx_schema = deconz_api.COMMAND_SCHEMAS[deconz_api.CommandId.aps_data_confirm]

        asyncio.get_running_loop().call_later(
            0.1,
            api.data_received,
            deconz_api.Command(
                command_id=deconz_api.CommandId.aps_data_confirm,
                seq=api._seq,
                payload=t.serialize_dict(
                    {
                        "status": deconz_api.Status.SUCCESS,
                        "frame_length": t.uint16_t(19),
                        "payload_length": t.uint16_t(12),
                        "device_state": deconz_api.DeviceState(
                            network_state=deconz_api.NetworkState2.CONNECTED,
                            device_state=(
                                deconz_api.DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
                                | deconz_api.DeviceStateFlags.APSDE_DATA_INDICATION
                            ),
                        ),
                        "request_id": t.uint8_t(16),
                        "dst_addr": t.DeconzAddressEndpoint.deserialize(
                            b"\x02\xaa\x55\x01"
                        )[0],
                        "src_ep": t.uint8_t(0),
                        "confirm_status": deconz_api.TXStatus.SUCCESS,
                        "reserved1": t.uint8_t(0),
                        "reserved2": t.uint8_t(0),
                        "reserved3": t.uint8_t(0),
                        "reserved4": t.uint8_t(0),
                    },
                    rx_schema,
                ),
            ).serialize(),
        )

        async with asyncio_timeout(0.2):
            rsp = await api.send_command(deconz_api.CommandId.aps_data_confirm)

    assert rsp["request_id"] == 16

    assert (
        "Firmware responded incorrectly (Response is mismatched! Sent"
        " <CommandId.aps_data_confirm: 4>, received <CommandId.version: 13>), retrying"
    ) not in caplog.text


def test_get_command_priority(api):
    assert (
        api._get_command_priority(
            deconz_api.Command(command_id=deconz_api.CommandId.write_parameter)
        )
        > api._get_command_priority(
            deconz_api.Command(command_id=deconz_api.CommandId.update_neighbor)
        )
        > api._get_command_priority(
            deconz_api.Command(command_id=deconz_api.CommandId.aps_data_request)
        )
    )



"""Data types module."""


import zigpy.types as zigpy_t
from zigpy.types import (  # noqa: F401
    EUI64,
    NWK,
    ExtendedPanId,
    IntStruct,
    LongOctetString,
    LVBytes,
    LVList,
    PanId,
    Struct,
    bitmap3,
    bitmap5,
    bitmap6,
    bitmap8,
    bitmap16,
    enum2,
    enum3,
    enum8,
    int8s,
    uint8_t,
    uint16_t,
    uint32_t,
    uint64_t,
)


def serialize_dict(data, schema):
    chunks = []

    for key in schema:
        value = data[key]
        if value is None:
            break

        if not isinstance(value, schema[key]):
            value = schema[key](value)

        chunks.append(value.serialize())

    return b"".join(chunks)


def deserialize_dict(data, schema):
    result = {}
    for name, type_ in schema.items():
        try:
            result[name], data = type_.deserialize(data)
        except ValueError:
            if data:
                raise

            result[name] = None
    return result, data


def list_replace(lst: list, old: object, new: object) -> list:
    """Replace all occurrences of `old` with `new` in `lst`."""
    return [new if x == old else x for x in lst]


class Bytes(bytes):
    def serialize(self):
        return self

    @classmethod
    def deserialize(cls, data):
        return cls(data), b""


class AddressMode(enum8):
    # Address modes used in deconz protocol

    GROUP = 0x01
    NWK = 0x02
    IEEE = 0x03
    NWK_AND_IEEE = 0x04


class DeconzSendDataFlags(bitmap8):
    NONE = 0x00
    NODE_ID = 0x01
    RELAYS = 0x02


class DeconzTransmitOptions(bitmap8):
    NONE = 0x00
    SECURITY_ENABLED = 0x01
    USE_NWK_KEY_SECURITY = 0x02
    USE_APS_ACKS = 0x04
    ALLOW_FRAGMENTATION = 0x08


class NWKList(LVList):
    _length_type = uint8_t
    _item_type = NWK


ZIGPY_ADDR_MODE_MAPPING = {
    zigpy_t.AddrMode.NWK: AddressMode.NWK,
    zigpy_t.AddrMode.IEEE: AddressMode.IEEE,
    zigpy_t.AddrMode.Group: AddressMode.GROUP,
    zigpy_t.AddrMode.Broadcast: AddressMode.NWK,
}


ZIGPY_ADDR_TYPE_MAPPING = {
    zigpy_t.AddrMode.NWK: NWK,
    zigpy_t.AddrMode.IEEE: EUI64,
    zigpy_t.AddrMode.Group: NWK,
    zigpy_t.AddrMode.Broadcast: NWK,
}


ZIGPY_ADDR_MODE_REVERSE_MAPPING = {
    AddressMode.NWK: zigpy_t.AddrMode.NWK,
    AddressMode.IEEE: zigpy_t.AddrMode.IEEE,
    AddressMode.GROUP: zigpy_t.AddrMode.Group,
    AddressMode.NWK_AND_IEEE: zigpy_t.AddrMode.IEEE,
}


ZIGPY_ADDR_TYPE_REVERSE_MAPPING = {
    AddressMode.NWK: zigpy_t.NWK,
    AddressMode.IEEE: zigpy_t.EUI64,
    AddressMode.GROUP: zigpy_t.Group,
    AddressMode.NWK_AND_IEEE: zigpy_t.NWK,
}


class DeconzAddress(Struct):
    address_mode: AddressMode
    address: EUI64
    ieee: EUI64

    @classmethod
    def deserialize(cls, data):
        r = cls()
        mode, data = AddressMode.deserialize(data)
        r.address_mode = mode
        if mode in [AddressMode.GROUP, AddressMode.NWK, AddressMode.NWK_AND_IEEE]:
            r.address, data = NWK.deserialize(data)
        elif mode == AddressMode.IEEE:
            r.address, data = EUI64.deserialize(data)
        if mode == AddressMode.NWK_AND_IEEE:
            r.ieee, data = EUI64.deserialize(data)
        return r, data

    def serialize(self):
        r = self.address_mode.serialize() + self.address.serialize()
        if self.address_mode == AddressMode.NWK_AND_IEEE:
            r += self.ieee.serialize()
        return r

    def as_zigpy_type(self):
        addr_mode = ZIGPY_ADDR_MODE_REVERSE_MAPPING[self.address_mode]
        address = ZIGPY_ADDR_TYPE_REVERSE_MAPPING[self.address_mode](self.address)

        if self.address_mode == AddressMode.NWK and self.address > 0xFFF7:
            addr_mode = zigpy_t.AddrMode.Broadcast
            address = zigpy_t.BroadcastAddress(self.address)
        elif self.address_mode == AddressMode.NWK_AND_IEEE:
            address = zigpy_t.EUI64(self.ieee)

        return zigpy_t.AddrModeAddress(
            addr_mode=addr_mode,
            address=address,
        )

    @classmethod
    def from_zigpy_type(cls, addr):
        instance = cls()
        instance.address_mode = ZIGPY_ADDR_MODE_MAPPING[addr.addr_mode]
        instance.address = ZIGPY_ADDR_TYPE_MAPPING[addr.addr_mode](addr.address)

        return instance


class DeconzAddressEndpoint(Struct):
    address_mode: AddressMode
    address: EUI64
    ieee: EUI64
    endpoint: uint8_t

    @classmethod
    def deserialize(cls, data):
        r, data = DeconzAddress.deserialize.__func__(cls, data)

        if r.address_mode in (
            AddressMode.NWK,
            AddressMode.IEEE,
            AddressMode.NWK_AND_IEEE,
        ):
            r.endpoint, data = uint8_t.deserialize(data)
        else:
            r.endpoint = None

        return r, data

    def serialize(self):
        r = uint8_t(self.address_mode).serialize()

        if self.address_mode in (AddressMode.NWK, AddressMode.NWK_AND_IEEE):
            r += NWK(self.address).serialize()
        elif self.address_mode == AddressMode.GROUP:
            r += NWK(self.address).serialize()

        if self.address_mode in (AddressMode.IEEE, AddressMode.NWK_AND_IEEE):
            r += EUI64(self.address).serialize()

        if self.address_mode in (
            AddressMode.NWK,
            AddressMode.IEEE,
            AddressMode.NWK_AND_IEEE,
        ):
            r += uint8_t(self.endpoint).serialize()

        return r

    @classmethod
    def from_zigpy_type(cls, addr, endpoint):
        temp_addr = DeconzAddress.from_zigpy_type(addr)

        instance = cls()
        instance.address_mode = temp_addr.address_mode
        instance.address = temp_addr.address
        instance.endpoint = endpoint

        return instance


class DataIndicationFlags(bitmap8):
    Always_Use_NWK_Source_Addr = 0b00000001
    Last_Hop_In_Reserved_Bytes = 0b00000010
    Include_Both_NWK_And_IEEE = 0b00000100



"""deCONZ serial protocol API."""

from __future__ import annotations

import asyncio
import collections
import itertools
import logging
import sys
from typing import Any, Callable

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout  # pragma: no cover
else:
    from asyncio import timeout as asyncio_timeout  # pragma: no cover

from zigpy.datastructures import PriorityLock
from zigpy.types import (
    APSStatus,
    Bool,
    Channels,
    KeyData,
    SerializableBytes,
    Struct,
    ZigbeePacket,
)
from zigpy.zdo.types import SimpleDescriptor

from zigpy_deconz.exception import CommandError, MismatchedResponseError, ParsingError
import zigpy_deconz.types as t
import zigpy_deconz.uart
from zigpy_deconz.utils import restart_forever

LOGGER = logging.getLogger(__name__)

MISMATCHED_RESPONSE_TIMEOUT = 0.5
COMMAND_TIMEOUT = 1.8
PROBE_TIMEOUT = 2

FRAME_LENGTH = object()
PAYLOAD_LENGTH = object()


class Status(t.enum8):
    SUCCESS = 0
    FAILURE = 1
    BUSY = 2
    TIMEOUT = 3
    UNSUPPORTED = 4
    ERROR = 5
    NO_NETWORK = 6
    INVALID_VALUE = 7


class NetworkState2(t.enum2):
    OFFLINE = 0
    JOINING = 1
    CONNECTED = 2
    LEAVING = 3


class DeviceStateFlags(t.bitmap6):
    APSDE_DATA_CONFIRM = 0b00001
    APSDE_DATA_INDICATION = 0b000010
    CONF_CHANGED = 0b000100
    APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE = 0b0001000


class DeviceState(t.Struct):
    network_state: NetworkState2
    device_state: DeviceStateFlags


class FirmwarePlatform(t.enum8):
    Conbee = 0x05
    Conbee_II = 0x07
    Conbee_III = 0x09


class FirmwareVersion(t.IntStruct, t.uint32_t):
    reserved: t.uint8_t
    platform: FirmwarePlatform
    minor: t.uint8_t
    major: t.uint8_t


class NetworkState(t.enum8):
    OFFLINE = 0
    JOINING = 1
    CONNECTED = 2
    LEAVING = 3


class SecurityMode(t.enum8):
    NO_SECURITY = 0x00
    PRECONFIGURED_NETWORK_KEY = 0x01
    NETWORK_KEY_FROM_TC = 0x02
    ONLY_TCLK = 0x03


class ZDPResponseHandling(t.bitmap16):
    NONE = 0x0000
    NodeDescRsp = 0x0001


class CommandId(t.enum8):
    aps_data_confirm = 0x04
    device_state = 0x07
    change_network_state = 0x08
    read_parameter = 0x0A
    write_parameter = 0x0B
    version = 0x0D
    device_state_changed = 0x0E
    aps_data_request = 0x12
    aps_data_indication = 0x17
    zigbee_green_power = 0x19
    mac_poll = 0x1C
    update_neighbor = 0x1D
    mac_beacon_indication = 0x1F


class TXStatus(t.enum8):
    SUCCESS = 0x00

    @classmethod
    def _missing_(cls, value):
        chained = APSStatus(value)
        status = t.uint8_t.__new__(cls, chained.value)
        status._name_ = chained.name
        status._value_ = value
        return status


class NetworkParameter(t.enum8):
    mac_address = 0x01
    nwk_panid = 0x05
    nwk_address = 0x07
    nwk_extended_panid = 0x08
    aps_designed_coordinator = 0x09
    channel_mask = 0x0A
    aps_extended_panid = 0x0B
    trust_center_address = 0x0E
    security_mode = 0x10
    configure_endpoint = 0x13
    use_predefined_nwk_panid = 0x15
    network_key = 0x18
    link_key = 0x19
    current_channel = 0x1C
    permit_join = 0x21
    protocol_version = 0x22
    nwk_update_id = 0x24
    watchdog_ttl = 0x26
    nwk_frame_counter = 0x27
    app_zdp_response_handling = 0x28


class IndexedKey(Struct):
    index: t.uint8_t
    key: KeyData


class LinkKey(Struct):
    ieee: t.EUI64
    key: KeyData


class IndexedEndpoint(Struct):
    index: t.uint8_t
    descriptor: SimpleDescriptor


class UpdateNeighborAction(t.enum8):
    ADD = 0x01


NETWORK_PARAMETER_TYPES = {
    NetworkParameter.mac_address: (None, t.EUI64),
    NetworkParameter.nwk_panid: (None, t.PanId),
    NetworkParameter.nwk_address: (None, t.NWK),
    NetworkParameter.nwk_extended_panid: (None, t.ExtendedPanId),
    NetworkParameter.aps_designed_coordinator: (None, t.uint8_t),
    NetworkParameter.channel_mask: (None, Channels),
    NetworkParameter.aps_extended_panid: (None, t.ExtendedPanId),
    NetworkParameter.trust_center_address: (None, t.EUI64),
    NetworkParameter.security_mode: (None, t.uint8_t),
    NetworkParameter.configure_endpoint: (t.uint8_t, IndexedEndpoint),
    NetworkParameter.use_predefined_nwk_panid: (None, Bool),
    NetworkParameter.network_key: (t.uint8_t, IndexedKey),
    NetworkParameter.link_key: (t.EUI64, LinkKey),
    NetworkParameter.current_channel: (None, t.uint8_t),
    NetworkParameter.permit_join: (None, t.uint8_t),
    NetworkParameter.protocol_version: (None, t.uint16_t),
    NetworkParameter.nwk_update_id: (None, t.uint8_t),
    NetworkParameter.watchdog_ttl: (None, t.uint32_t),
    NetworkParameter.nwk_frame_counter: (None, t.uint32_t),
    NetworkParameter.app_zdp_response_handling: (None, ZDPResponseHandling),
}


class Command(Struct):
    command_id: CommandId
    seq: t.uint8_t
    payload: t.Bytes


COMMAND_SCHEMAS = {
    CommandId.update_neighbor: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
            "action": UpdateNeighborAction,
            "nwk": t.NWK,
            "ieee": t.EUI64,
            "mac_capability_flags": t.uint8_t,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "action": UpdateNeighborAction,
            "nwk": t.NWK,
            "ieee": t.EUI64,
            "mac_capability_flags": t.uint8_t,
        },
    ),
    CommandId.aps_data_confirm: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "device_state": DeviceState,
            "request_id": t.uint8_t,
            "dst_addr": t.DeconzAddressEndpoint,
            "src_ep": t.uint8_t,
            "confirm_status": TXStatus,
            "reserved1": t.uint8_t,
            "reserved2": t.uint8_t,
            "reserved3": t.uint8_t,
            "reserved4": t.uint8_t,
        },
    ),
    CommandId.aps_data_indication: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
            "flags": t.DataIndicationFlags,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "device_state": DeviceState,
            "dst_addr": t.DeconzAddress,
            "dst_ep": t.uint8_t,
            "src_addr": t.DeconzAddress,
            "src_ep": t.uint8_t,
            "profile_id": t.uint16_t,
            "cluster_id": t.uint16_t,
            "asdu": t.LongOctetString,
            "reserved1": t.uint8_t,
            "reserved2": t.uint8_t,
            "lqi": t.uint8_t,
            "reserved3": t.uint8_t,
            "reserved4": t.uint8_t,
            "reserved5": t.uint8_t,
            "reserved6": t.uint8_t,
            "rssi": t.int8s,
        },
    ),
    CommandId.aps_data_request: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
            "request_id": t.uint8_t,
            "flags": t.DeconzSendDataFlags,
            "dst": t.DeconzAddressEndpoint,
            "profile_id": t.uint16_t,
            "cluster_id": t.uint16_t,
            "src_ep": t.uint8_t,
            "asdu": t.LongOctetString,
            "tx_options": t.DeconzTransmitOptions,
            "radius": t.uint8_t,
            "relays": t.NWKList,  # optional
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "device_state": DeviceState,
            "request_id": t.uint8_t,
        },
    ),
    CommandId.change_network_state: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            # "payload_length": PAYLOAD_LENGTH,
            "network_state": NetworkState,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            # "payload_length": t.uint16_t,
            "network_state": NetworkState,
        },
    ),
    CommandId.device_state: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            # "payload_length": PAYLOAD_LENGTH,
            "reserved1": t.uint8_t(0),
            "reserved2": t.uint8_t(0),
            "reserved3": t.uint8_t(0),
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            # "payload_length": t.uint16_t,
            "device_state": DeviceState,
            "reserved1": t.uint8_t,
            "reserved2": t.uint8_t,
        },
    ),
    CommandId.device_state_changed: (
        None,
        {
            "status": Status,
            "frame_length": t.uint16_t,
            # "payload_length": t.uint16_t,
            "device_state": DeviceState,
            "reserved": t.uint8_t,
        },
    ),
    CommandId.mac_poll: (
        None,
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "src_addr": t.DeconzAddress,
            "lqi": t.uint8_t,
            "rssi": t.int8s,
            "life_time": t.uint32_t,  # Optional
            "device_timeout": t.uint32_t,  # Optional
        },
    ),
    CommandId.read_parameter: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
            "parameter_id": NetworkParameter,
            "parameter": t.Bytes,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "parameter_id": NetworkParameter,
            "parameter": t.Bytes,
        },
    ),
    CommandId.version: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            # "payload_length": PAYLOAD_LENGTH,
            "reserved": t.uint32_t(0),
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            # "payload_length": t.uint16_t,
            "version": FirmwareVersion,
        },
    ),
    CommandId.write_parameter: (
        {
            "status": Status.SUCCESS,
            "frame_length": FRAME_LENGTH,
            "payload_length": PAYLOAD_LENGTH,
            "parameter_id": NetworkParameter,
            "parameter": t.Bytes,
        },
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "parameter_id": NetworkParameter,
        },
    ),
    CommandId.zigbee_green_power: (
        None,
        {
            "status": Status,
            "frame_length": t.uint16_t,
            "payload_length": t.uint16_t,
            "reserved": t.LongOctetString,
        },
    ),
}


class Deconz:
    """deCONZ API class."""

    def __init__(self, app: Callable, device_config: dict[str, Any]):
        """Init instance."""
        self._app = app

        # [seq][cmd_id] = [fut1, fut2, ...]
        self._awaiting = collections.defaultdict(lambda: collections.defaultdict(list))
        self._mismatched_response_timers: dict[int, asyncio.TimerHandle] = {}
        self._command_lock = PriorityLock()
        self._config = device_config
        self._device_state = DeviceState(
            network_state=NetworkState2.OFFLINE,
            device_state=(
                DeviceStateFlags.APSDE_DATA_CONFIRM
                | DeviceStateFlags.APSDE_DATA_INDICATION
            ),
        )

        self._free_slots_available_event = asyncio.Event()
        self._free_slots_available_event.set()

        self._data_poller_event = asyncio.Event()
        self._data_poller_event.set()
        self._data_poller_task: asyncio.Task | None = None

        self._seq = 1
        self._protocol_version = 0
        self._firmware_version = FirmwareVersion(0)
        self._uart: zigpy_deconz.uart.Gateway | None = None

    @property
    def firmware_version(self) -> FirmwareVersion:
        """Return ConBee firmware version."""
        return self._firmware_version

    @property
    def network_state(self) -> NetworkState:
        """Return current network state."""
        return self._device_state.network_state

    @property
    def protocol_version(self) -> int:
        """Protocol Version."""
        return self._protocol_version

    async def connect(self) -> None:
        assert self._uart is None

        self._uart = await zigpy_deconz.uart.connect(self._config, self)

        try:
            await self.version()
            device_state_rsp = await self.send_command(CommandId.device_state)
        except Exception:
            await self.disconnect()
            self._uart = None
            raise

        self._device_state = device_state_rsp["device_state"]

        self._data_poller_task = asyncio.create_task(self._data_poller())

    def connection_lost(self, exc: Exception | None) -> None:
        """Lost serial connection."""
        if self._app is not None:
            self._app.connection_lost(exc)

    async def disconnect(self):
        if self._data_poller_task is not None:
            self._data_poller_task.cancel()
            self._data_poller_task = None

        if self._uart is not None:
            await self._uart.disconnect()
            self._uart = None

        self._app = None

    def _get_command_priority(self, command: Command) -> int:
        return {
            # The watchdog is fed using `write_parameter` and `get_device_state` so they
            # must take priority
            CommandId.write_parameter: 999,
            CommandId.device_state: 999,
            # APS data requests are retried and can be deprioritized
            CommandId.aps_data_request: -1,
        }.get(command.command_id, 0)

    async def send_command(self, cmd, **kwargs) -> Any:
        while True:
            try:
                return await self._command(cmd, **kwargs)
            except MismatchedResponseError as exc:
                LOGGER.debug("Firmware responded incorrectly (%s), retrying", exc)

    async def _command(self, cmd, **kwargs):
        payload = []
        tx_schema, _ = COMMAND_SCHEMAS[cmd]
        trailing_optional = False

        for name, param_type in tx_schema.items():
            if isinstance(param_type, int):
                if name not in kwargs:
                    # Default value
                    value = param_type.serialize()
                else:
                    value = type(param_type)(kwargs[name]).serialize()
            elif name in ("frame_length", "payload_length"):
                value = param_type
            elif kwargs.get(name) is None:
                trailing_optional = True
                value = None
            elif not isinstance(kwargs[name], param_type):
                value = param_type(kwargs[name]).serialize()
            else:
                value = kwargs[name].serialize()

            if value is None:
                continue

            if trailing_optional:
                raise ValueError(
                    f"Command {cmd} with kwargs {kwargs}"
                    f" has non-trailing optional argument"
                )

            payload.append(value)

        if PAYLOAD_LENGTH in payload:
            payload = t.list_replace(
                lst=payload,
                old=PAYLOAD_LENGTH,
                new=t.uint16_t(
                    sum(len(p) for p in payload[payload.index(PAYLOAD_LENGTH) + 1 :])
                ).serialize(),
            )

        if FRAME_LENGTH in payload:
            payload = t.list_replace(
                lst=payload,
                old=FRAME_LENGTH,
                new=t.uint16_t(
                    2 + sum(len(p) if p is not FRAME_LENGTH else 2 for p in payload)
                ).serialize(),
            )

        command = Command(
            command_id=cmd,
            seq=None,
            payload=b"".join(payload),
        )

        if self._uart is None:
            # connection was lost
            raise CommandError(
                "API is not running",
                status=Status.ERROR,
                command=command,
            )

        async with self._command_lock(priority=self._get_command_priority(command)):
            seq = self._seq
            self._seq = (self._seq % 255) + 1

            fut = asyncio.Future()
            self._awaiting[seq][cmd].append(fut)

            try:
                LOGGER.debug("Sending %s%s (seq=%s)", cmd, kwargs, seq)
                self._uart.send(command.replace(seq=seq).serialize())

                async with asyncio_timeout(COMMAND_TIMEOUT):
                    return await fut
            except asyncio.TimeoutError:
                LOGGER.debug("No response to '%s' command with seq %d", cmd, seq)
                raise
            finally:
                self._awaiting[seq][cmd].remove(fut)

    def data_received(self, data: bytes) -> None:
        command, _ = Command.deserialize(data)

        if command.command_id not in COMMAND_SCHEMAS:
            LOGGER.warning("Unknown command received: %s", command)
            return

        _, rx_schema = COMMAND_SCHEMAS[command.command_id]

        fut = None
        wrong_fut_cmd_id = None

        try:
            fut = self._awaiting[command.seq][command.command_id][0]
        except IndexError:
            # XXX: The firmware can sometimes respond with the wrong response. Find the
            # future associated with it so we can throw an appropriate error.
            for cmd_id, futs in self._awaiting[command.seq].items():
                if futs:
                    fut = futs[0]
                    wrong_fut_cmd_id = cmd_id
                    break

        try:
            params, rest = t.deserialize_dict(command.payload, rx_schema)
        except Exception:
            LOGGER.debug("Failed to parse command %s", command, exc_info=True)

            if fut is not None and not fut.done():
                fut.set_exception(
                    ParsingError(
                        f"Failed to parse command: {command}",
                        status=Status.ERROR,
                        command=command,
                    )
                )

            return

        if rest:
            LOGGER.debug("Unparsed data remains after frame: %s, %s", command, rest)

        assert params["frame_length"] == len(data)

        if "payload_length" in params:
            running_length = itertools.accumulate(
                len(v.serialize()) if v is not None else 0 for v in params.values()
            )
            length_at_param = dict(zip(params.keys(), running_length))

            assert (
                len(data) - length_at_param["payload_length"] - 2
                == params["payload_length"]
            )

        LOGGER.debug(
            "Received command %s%s (seq %d)", command.command_id, params, command.seq
        )
        status = params["status"]

        exc = None

        # Make sure to clear any pending mismatched response timers
        if command.seq in self._mismatched_response_timers:
            LOGGER.debug("Clearing existing mismatched response timer")
            self._mismatched_response_timers.pop(command.seq).cancel()

        if wrong_fut_cmd_id is not None:
            LOGGER.debug(
                "Mismatched response, triggering error in %0.2fs",
                MISMATCHED_RESPONSE_TIMEOUT,
            )
            # The firmware *sometimes* responds with the correct response later
            self._mismatched_response_timers[
                command.seq
            ] = asyncio.get_event_loop().call_later(
                MISMATCHED_RESPONSE_TIMEOUT,
                fut.set_exception,
                MismatchedResponseError(
                    command.command_id,
                    params,
                    (
                        f"Response is mismatched! Sent {wrong_fut_cmd_id},"
                        f" received {command.command_id}"
                    ),
                ),
            )

            # Make sure we do not resolve the future
            fut = None
        elif status != Status.SUCCESS:
            exc = CommandError(
                f"{command.command_id}, status: {status}",
                status=status,
                command=command,
            )

        if fut is not None:
            try:
                if exc is None:
                    fut.set_result(params)
                else:
                    fut.set_exception(exc)
            except asyncio.InvalidStateError:
                LOGGER.debug(
                    "Duplicate or delayed response for seq %s (awaiting %s)",
                    command.seq,
                    self._awaiting[command.seq],
                )

            if exc is not None:
                return

        if handler := getattr(self, f"_handle_{command.command_id.name}", None):
            handler_params = {
                k: v
                for k, v in params.items()
                if k not in ("frame_length", "payload_length")
            }

            # Queue up the callback within the event loop
            asyncio.get_running_loop().call_soon(lambda: handler(**handler_params))

    @restart_forever
    async def _data_poller(self):
        while True:
            await self._data_poller_event.wait()
            self._data_poller_event.clear()

            if self._device_state.network_state == NetworkState2.OFFLINE:
                continue

            # Poll data indication
            if (
                DeviceStateFlags.APSDE_DATA_INDICATION
                in self._device_state.device_state
            ):
                # Old Conbee I firmware has an addressing bug for incoming multicasts
                if (
                    self.protocol_version >= 0x010B
                    and self.firmware_version.platform == FirmwarePlatform.Conbee
                ):
                    flags = t.DataIndicationFlags.Include_Both_NWK_And_IEEE
                else:
                    flags = t.DataIndicationFlags.Always_Use_NWK_Source_Addr

                rsp = await self.send_command(
                    CommandId.aps_data_indication, flags=flags
                )
                self._handle_device_state_changed(
                    status=rsp["status"], device_state=rsp["device_state"]
                )

                self._app.packet_received(
                    ZigbeePacket(
                        src=rsp["src_addr"].as_zigpy_type(),
                        src_ep=rsp["src_ep"],
                        dst=rsp["dst_addr"].as_zigpy_type(),
                        dst_ep=rsp["dst_ep"],
                        tsn=None,
                        profile_id=rsp["profile_id"],
                        cluster_id=rsp["cluster_id"],
                        data=SerializableBytes(rsp["asdu"]),
                        lqi=rsp["lqi"],
                        rssi=rsp["rssi"],
                    )
                )

            # Poll data confirm
            if DeviceStateFlags.APSDE_DATA_CONFIRM in self._device_state.device_state:
                rsp = await self.send_command(CommandId.aps_data_confirm)

                self._app.handle_tx_confirm(rsp["request_id"], rsp["confirm_status"])
                self._handle_device_state_changed(
                    status=rsp["status"], device_state=rsp["device_state"]
                )

    def _handle_device_state_changed(
        self,
        status: t.Status,
        device_state: DeviceState,
        reserved: t.uint8_t = 0,
    ) -> None:
        if device_state.network_state != self.network_state:
            LOGGER.debug(
                "Network device_state transition: %s -> %s",
                self.network_state.name,
                device_state.network_state.name,
            )

        if (
            DeviceStateFlags.APSDE_DATA_REQUEST_FREE_SLOTS_AVAILABLE
            in device_state.device_state
        ):
            self._free_slots_available_event.set()
        else:
            self._free_slots_available_event.clear()

        self._device_state = device_state
        self._data_poller_event.set()

    def _handle_device_state(
        self,
        status: t.Status,
        device_state: DeviceState,
        reserved1: t.uint8_t,
        reserved2: t.uint8_t,
    ) -> None:
        if (
            self.firmware_version.platform == FirmwarePlatform.Conbee_III
            and self.firmware_version == 0x26450900
        ):
            # Initial Conbee III firmware used the wrong command to notify of network
            # state changes
            self._handle_device_state_changed(status=status, device_state=device_state)

    async def version(self):
        self._protocol_version = await self.read_parameter(
            NetworkParameter.protocol_version
        )

        version_rsp = await self.send_command(CommandId.version, reserved=0)
        self._firmware_version = version_rsp["version"]

        return self.firmware_version

    async def read_parameter(
        self, parameter_id: NetworkParameter, parameter: Any = None
    ) -> Any:
        read_param_type, write_param_type = NETWORK_PARAMETER_TYPES[parameter_id]

        if parameter is None:
            value = t.Bytes(b"")
        else:
            value = read_param_type(parameter).serialize()

        rsp = await self.send_command(
            CommandId.read_parameter,
            parameter_id=parameter_id,
            parameter=value,
        )

        assert rsp["parameter_id"] == parameter_id

        result, _ = write_param_type.deserialize(rsp["parameter"])
        LOGGER.debug("Read parameter %s(%s)=%r", parameter_id.name, parameter, result)

        return result

    async def write_parameter(
        self, parameter_id: NetworkParameter, parameter: Any
    ) -> None:
        read_param_type, write_param_type = NETWORK_PARAMETER_TYPES[parameter_id]
        await self.send_command(
            CommandId.write_parameter,
            parameter_id=parameter_id,
            parameter=write_param_type(parameter).serialize(),
        )

    async def aps_data_request(
        self,
        req_id,
        dst_addr_ep,
        profile,
        cluster,
        src_ep,
        aps_payload,
        *,
        relays=None,
        tx_options=t.DeconzTransmitOptions.USE_NWK_KEY_SECURITY,
        radius=0,
    ) -> None:
        flags = t.DeconzSendDataFlags.NONE

        # https://github.com/zigpy/zigpy-deconz/issues/180#issuecomment-1017932865
        if relays is not None:
            # There is a max of 9 relays
            assert len(relays) <= 9
            flags |= t.DeconzSendDataFlags.RELAYS

        if not self._free_slots_available_event.is_set():
            LOGGER.debug("Waiting for free slots to become available")
            await self._free_slots_available_event.wait()

        rsp = await self.send_command(
            CommandId.aps_data_request,
            request_id=req_id,
            flags=flags,
            dst=dst_addr_ep,
            profile_id=profile,
            cluster_id=cluster,
            src_ep=src_ep,
            asdu=aps_payload,
            tx_options=tx_options,
            radius=radius,
            relays=relays,
        )

        self._handle_device_state_changed(
            status=rsp["status"], device_state=rsp["device_state"]
        )

    async def get_device_state(self) -> DeviceState:
        rsp = await self.send_command(CommandId.device_state)

        return rsp["device_state"]

    async def change_network_state(self, new_state: NetworkState) -> None:
        await self.send_command(CommandId.change_network_state, network_state=new_state)

    async def add_neighbour(
        self, nwk: t.NWK, ieee: t.EUI64, mac_capability_flags: t.uint8_t
    ) -> None:
        try:
            await self.send_command(
                CommandId.update_neighbor,
                action=UpdateNeighborAction.ADD,
                nwk=nwk,
                ieee=ieee,
                mac_capability_flags=mac_capability_flags,
            )
        except ParsingError as exc:
            # Older Conbee III firmwares send back an invalid response
            status = Status(exc.command.payload[0])

            if status != Status.SUCCESS:
                raise CommandError(
                    f"{exc.command.command_id}, status: {status}",
                    status=status,
                    command=exc.command,
                ) from exc
